"""# RaGraph plotting module"""
from typing import Any, Dict, Iterable, List, Optional

from plotly import graph_objects as go

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node
from ragraph.plot.generic import Style

__all__ = ["Style", "dsm", "dmm", "mdm"]


def dsm(
    leafs: List[Node],
    edges: List[Edge],
    style: Style = Style(),
    sort: bool = True,
    sort_args: Dict[str, Any] = dict(),
    node_kind: Optional[str] = None,
    show: bool = False,
) -> go.Figure:
    """Get a DSM plot of a Graph object.

    Arguments
        leafs: The nodes to be placed on the rows and columns of the matrix..
        edges: Edges to be displayed between leaf nodes.
        style: Plot style option mapping.
        sort: Boolean to indicate whether the rows and cols should be sorted following
            the hierarchical structure.
        node_kind: The node kind to be displayed.
        show: Boolean to display the figure.
    """
    node_kinds = [node_kind] if node_kind else list({n.kind for n in leafs})
    assert (
        len(node_kinds) == 1
    ), "A DSM should only contain one node kind. See MDM for multi-domain matrices."

    return mdm(
        leafs,
        edges,
        style=style,
        sort=sort,
        sort_args=sort_args,
        node_kinds=node_kinds,
        show=show,
    )


def dmm(
    rows: List[Node],
    cols: List[Node],
    edges: List[Edge],
    style: Style = Style(),
    sort: bool = True,
    row_node_kinds: Optional[List[str]] = None,
    row_sort_args: Dict[str, Any] = dict(),
    col_node_kinds: Optional[List[str]] = None,
    col_sort_args: Dict[str, Any] = dict(),
    show: bool = False,
) -> go.Figure:
    """Get a domain-mapping-matrix (DMM) plot of a Graph object.

    Arguments:
        rows: The nodes to be placed on the rows of the matrix.
        cols: The columns to be placed on the columns of the matrix.
        edges: Edges to be displayed between leaf nodes.
        style: Plot style option mapping.
        sort: Boolean to indicate whether the rows and cols should be sorted following
            the hierarchical structure.
        row_node_kinds: The node kinds displayed on the rows.
        col_node_kinds: The node kinds displayed on the columns.
        show: Boolean to display the figure.

    Returns:
       Domain-mapping matrix figure.
    """
    from ragraph.plot import utils

    if sort:
        rows = utils.get_axis_sequence(rows, kinds=row_node_kinds, **row_sort_args)
        cols = utils.get_axis_sequence(cols, kinds=col_node_kinds, **col_sort_args)

    grid = utils.get_dmm_grid(rows=rows, cols=cols, edges=edges, style=style)

    fig = utils.get_subplots(grid, style=style)
    return utils.process_fig(fig=fig, show=show, style=style)


def mdm(
    leafs: List[Node],
    edges: List[Edge],
    style: Style = Style(),
    sort: bool = True,
    sort_args: Dict[str, Any] = dict(),
    node_kinds: Optional[List[str]] = None,
    show: bool = False,
) -> go.Figure:
    """Get a Multi-Domain Matrix (MDM) plot of a Graph object.

    Arguments
        leafs: The nodes to be placed on the rows and columns of the matrix..
        edges: Edges to be displayed between leaf nodes.
        style: Plot style option mapping.
        sort: Boolean to indicate whether the rows and cols should be sorted following
            the hierarchical structure.
        node_kinds: The node kinds displayed within the matrix.
        show: Boolean to display the figure.
    """
    from ragraph.plot import utils

    if sort:
        leafs = utils.get_axis_sequence(leafs, kinds=node_kinds, **sort_args)

    grid = utils.get_mdm_grid(leafs=leafs, edges=edges, style=style)

    fig = utils.get_subplots(grid, style=style)

    return utils.process_fig(fig=fig, show=show, style=style)


def delta_dsm(g1: Graph, g2: Graph, *args, style: Style = Style(), **kwargs) -> go.Figure:
    """Get a delta-DSM plot between two Graph objects."""
    raise NotImplementedError()


def sigma_dsm(graphs: Iterable[Graph], *args, style: Style = Style(), **kwargs) -> go.Figure:
    """Get a sigma-DSM plot of any number of Graph objects."""
    raise NotImplementedError()


def network(g: Graph, *args, style: Style = Style(), **kwargs) -> go.Figure:
    """Get a network plot of a Graph object."""
    raise NotImplementedError()



"""# Plot components.

This module contains several re-usable plot components that contain the data for a Plotly (sub-)
figure. See [`Component`][ragraph.plot.generic.Component] for the definition of a plot component.
"""

from ragraph.plot.components.blank import Blank
from ragraph.plot.components.labels import Labels
from ragraph.plot.components.legend import Legend
from ragraph.plot.components.piemap import PieMap
from ragraph.plot.components.tree import Tree

__all__ = ["Blank", "Labels", "Legend", "PieMap", "Tree"]



"""# Ford climate control system

This dataset describes a climate control system as it was to be found in Ford vehicles. Four
different dependency types have been documented, being spatial, energy flow, information flow, and
material flow dependencies. These are weighted from -2 to 2, where the following definitions have
been used:

    * +2: Required
    * +1: Desired
    *  0: Indifferent
    * -1: Undesired
    * -2: Detrimental

We have added an "adjacency" weight, which is the nonnegative sum of all dependencies between
components.

Reference:
    Pimmler, T. U., & Eppinger, S. D. (1994). Integration Analysis of Product Decompositions. ASME
    Design Theory and Methodology Conference.
"""


edge_weights = [
    "adjacency",
    "spatial",
    "energy flow",
    "information flow",
    "material flow",
]



"""
# Heuristics

Heuristics are combinations of algorithms available in the other sub-packages or have an output that
does not strictly fit one of the other categories. Not all heuristics have a common argument
structure because of their more diverse nature.

## Available heuristics

- [`johnson`][ragraph.analysis.heuristics.johnson] circuit finding algorithm.
- [`markov_gamma`][ragraph.analysis.heuristics.markov_gamma] hierarchical clustering with bus
  detection heuristic.
"""

from ragraph.analysis.heuristics._johnson import johnson
from ragraph.analysis.heuristics._markov_gamma import markov_gamma

__all__ = ["johnson", "markov_gamma"]



"""# Pre-clustered Ford climate control system using the Markov-Gamma heuristic

Pre-clustered using bus detection plus hierarchical clustering.

Reference:
    Pimmler, T. U., & Eppinger, S. D. (1994). Integration Analysis of Product Decompositions. ASME
    Design Theory and Methodology Conference.
"""


edge_weights = [
    "adjacency",
    "spatial",
    "energy flow",
    "information flow",
    "material flow",
]



"""Generic purpose classes and methods such as metadata models.

Metadata includes the kinds, labels, weights and annotations that we assign to our data
objects. Standardizing this enables a more predictable experience.
"""
import uuid as _uuid
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Set, Union
from warnings import warn


class MappingValidationError(Exception):
    pass


def field(fget: Callable):
    """A [`Mapping`][ragraph.generic.Mapping] `field` is a property that utilizes the
    [`Mapping`][ragraph.generic.Mapping]'s data.

    Use it like the ``@property`` decorator and leave the function body blank (or pass):

    By inspecting the wrapped method's name we derive the property's key.

    Getting data from the mapping is done by retrieving the key from the set values. If that key is
    not found, an attempt is made on the defaults. An error is thrown when the key is neither to be
    found in the data or defaults.

    Setting data checks for nested [`Mapping`][ragraph.generic.Mapping] keys to update those and
    otherwise defaults to simply storing the value. Updating a nested
    [`Mapping`][ragraph.generic.Mapping] field only updates the keys that are provided.

    Deleting data pops the key from the set values, but does not remove any defaults.

    Note:
        The method contents of a [`Mapping`][ragraph.generic.Mapping] `field` are ignored.
    """
    key = fget.__name__

    def getter(self):
        """Get a mapping value."""
        if key in self._data:
            return self._data[key]

        if key in self._defaults:
            return self._defaults[key]

        raise AttributeError(f"Mapping does not contain a (default) value for key '{key}'.")

    def setter(self, value: Any):
        """Set a mapping value."""
        # Setting the mapping to None means resetting it to default.
        if value is None:
            self._data.pop(key, None)
            return

        # Handle nested mapping keys.
        if self._is_mapping(key):
            if key in self._data:
                # Mapping already set, update it.
                self._data[key].update(value)
            else:
                # Create a copy of the default and update that.
                self._data[key] = deepcopy(self._defaults.get(key))
                self._data[key].update(value)
            return

        # Otherwise just store the value.
        else:
            self._data[key] = value

    def deleter(self):
        """Delete a mapping value (e.g. unset it)."""
        self._data.pop(key, None)

    return property(fget=getter, fset=setter, fdel=deleter, doc=fget.__doc__)


class Mapping:
    """A dictionary like object that with property-based access fields.

    It's possible to include allowed keys, default values, and optional validators for
    certain keys/properties of a derived class.

    ```py
    from ragraph.generic import Mapping, field
    def check_int(value):
        assert value == 1

    class MyMap(Mapping):
        _protected = True
        _defaults = dict(myfield=1)
        _validators = dict(myfield=check_int)
        @field
        def myfield(self) -> int:
            '''My field's docstring'''

    m = MyMap(myfield=3)
    assert m.myfield == 3,            "This should return our set value."
    assert m.myfield == m['myfield'], "A mapping works like a dictionary."
    del(m.myfield)                    # This deletes our override.
    m.validate()                      # Checks whether myfield == 1.
    ```
    """

    _protected = False
    """Setting public properties is restricted to those in
    [`self.keys`][ragraph.generic.Mapping.keys].
    """

    _defaults: Dict[str, Any] = dict()
    """Default values for included keys."""

    _keys: Optional[Set[str]] = None
    """Set of always allowed keys to set."""

    _validators: Dict[str, Callable] = dict()
    """Validators for included keys. Used when calling
    [`self.validate`][ragraph.generic.Mapping.validate].
    Also see [`self._post_validation`][ragraph.generic.Mapping._post_validation]."""

    def __init__(self, *args, **kwargs):
        # The internal storage dictionary.
        self._data: Dict[str, Any] = dict()

        # Update the mapping with args and kwargs.
        self.update(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({str(self.as_dict())})"

    def __repr__(self) -> str:
        values = ", ".join([str(k) + "=" + str(v) for k, v in self.as_dict().items()])
        return "<{}.{}({}) at {}>".format(
            self.__class__.__module__, self.__class__.__name__, values, hex(id(self))
        )

    def __eq__(self, other: Any) -> bool:
        try:
            return self.as_dict() == other.as_dict()
        except Exception:
            return False

    def __getitem__(self, key: str) -> Any:
        """Get an attribute via a dictionary accessor: `self['key']`."""
        return getattr(self, key)

    def get(self, key: str, fallback: Any = None) -> Any:
        """Get an attribute with a fallback value if it isn't found."""
        return getattr(self, key, fallback)

    def __setattr__(self, key: str, value: Any) -> None:
        """Set an attribute : `self.key = value`.

        Private properties are left untouched, public properties need permission.
        """
        if not key.startswith("_"):
            keys = set() if self._keys is None else self._keys
            keys = keys.union(self.keys())
            if key not in keys:
                if self._protected:
                    raise KeyError(f"Key or property '{key}' is not allowed. Only '{keys}'.")
                else:
                    # Dynamically create a new field.
                    def stub(self):
                        f"""{key} mapping field."""
                        pass

                    stub.__name__ = key
                    setattr(self.__class__, key, field(stub))

        super().__setattr__(key, value)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an attribute via a dictionary accessor: `self['key'] = value`.

        Private properties are left untouched, public properties need permission.
        """
        setattr(self, key, value)

    def update(self, *args: Dict[str, Any], **kwargs: Any) -> None:
        """Update multiple keys at once.

        Arguments:
            args: Dictionaries of key value pairs to update the set data with.
            kwargs: Keyword arguments to update the set data with.
        """
        for arg in args:
            try:
                for key, value in arg.items():
                    setattr(self, key, value)
            except AttributeError:
                raise TypeError(f"Expected a dictionary or a Mapping. Found a {type(arg)}.")
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __iter__(self):
        """Provide an iterator."""
        yield from self.items()

    def _is_mapping(self, key) -> bool:
        """Check whether this key is a mapping."""
        return isinstance(getattr(self, key, None), Mapping)

    def items(self):
        """Get defaults and overrides as a (key, value) dict_items iterator."""
        m = deepcopy(self._defaults)
        m.update(self._data)
        return m.items()

    def keys(self) -> Set[str]:
        """The keys in this mapping. (defaults and overrides)."""
        return set(self._defaults.keys()).union(set(self._data.keys()))

    def validate(self) -> None:
        """Check whether the current data passes validation."""
        # Get the complete mapping (default overridden with set values).
        m = deepcopy(self._defaults)
        m.update(self._data)

        # Per key validation.
        for key, value in m.items():
            try:
                # Validate submappings.
                if isinstance(value, Mapping):
                    value.validate()

                # Validate using validator if available.
                validator = self._validators.get(key)
                if validator is not None:
                    validator(value)
            except Exception:
                raise MappingValidationError(
                    f"Validation of '{key}' failed: '{value}' did not pass validation."
                )

        # Post-validation.
        try:
            self._post_validation()
        except Exception:
            raise MappingValidationError("Post-validation did not succeed.")

    def _post_validation(self) -> None:
        """Validation to run at the end of regular validation. Does nothing by default.

        Useful when a derived class needs to cross-validate certain keys instead of just
        a dedicated method per key.
        """
        pass

    def as_dict(self) -> Dict[str, Any]:
        """Return a copy as a dictionary with all sub-mappings as dictionaries, too."""
        m = self._defaults.copy()
        m.update(self._data)
        for key, value in m.items():
            if isinstance(value, Mapping):
                m[key] = value.as_dict()
        return deepcopy(m)


class Annotations(Mapping):
    """Miscellaneous properties mapping such as tool-specific metadata."""

    _protected = False


class Bound:
    """Numerical lower or upper bound for a value. Use with comparison operators.

    Arguments:
        value: Numerical bound value.
        inclusive: Whether the bound is inclusive.
        report: Whether to report an "error", "warn" or nothing (`None`) on
            bound violations.
    """

    def __init__(
        self,
        value: Union[int, float],
        inclusive: bool = True,
        report: Optional["str"] = None,
    ):
        self.value = value
        self.inclusive = inclusive
        self.report = report

    def __str__(self) -> str:
        return "Bound(value={}, inclusive={}, report={})".format(
            self.value, self.inclusive, self.report
        )

    def __lt__(self, other: Union[int, float]) -> bool:
        """Used as a lower bound."""
        if other < self.value:
            if self.report is not None:
                self._report(f"Value '{other}' below bound '{self.value}'.")
            return False
        elif not self.inclusive and other == self.value:
            if self.report is not None:
                self._report(f"Value '{other}' is at the exclusive lower bound value.")
            return False
        return True

    def __gt__(self, other: Union[int, float]) -> bool:
        """Used as an upper bound."""
        if other > self.value:
            if self.report is not None:
                self._report(f"Value '{other}' above bound '{self.value}'.")
            return False
        elif not self.inclusive and other == self.value:
            if self.report is not None:
                self._report(f"Value '{other}' is at the exclusive upper bound value.")
            return False
        return True

    def _report(self, msg: str) -> None:
        """Report an error or warning depending on if the bound is set to 'hard'."""
        if self.report == "warn":
            warn(msg)
        else:
            raise ValueError(msg)

    def as_dict(self) -> Dict[str, Any]:
        """Serializable dictionary representation."""
        return dict(value=self.value, inclusive=self.inclusive, report=self.report)


class ContinuousDomain:
    """Numerical domain for a value. Use with "in" operator."""

    def __init__(
        self,
        lower: Optional[Bound] = None,
        upper: Optional[Bound] = None,
    ):
        self.lower = lower
        self.upper = upper

    def __str__(self) -> str:
        if self.lower is None:
            lb = "(-inf"
        else:
            lb = "{}{}".format("[" if self.lower.inclusive else "(", self.lower.value)

        if self.upper is None:
            ub = "inf)"
        else:
            ub = "{}{}".format(self.upper.value, "]" if self.upper.inclusive else "]")
        return f"{lb}, {ub}"

    def __repr__(self) -> str:
        return "<{}.{}({}) at {}>".format(
            self.__class__.__module__, self.__class__.__name__, str(self), hex(id(self))
        )

    def __contains__(self, value: Union[int, float]) -> bool:
        """Whether this continuous domain contains a value."""
        return (value > self.lower if self.lower is not None else True) and (
            value < self.upper if self.upper is not None else True
        )

    def as_dict(self) -> Dict[str, Any]:
        """Serializable dictionary representation."""
        return dict(
            lower=None if self.lower is None else self.lower.as_dict(),
            upper=None if self.upper is None else self.upper.as_dict(),
        )


class Metadata:
    """Metadata for graph elements.

    Arguments:
        name: Instance name. Set to a copy of the UUID if `None` provided.
        kind: Kind or main category of this instance.
        labels: Labels categorizing this instance.
        weights: Dictionary of weights attached to this instance.
        annotations: Miscellaneous properties of this instance.
        uuid: UUID of this instance, generated when `None` provided.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        kind: Optional[str] = None,
        labels: Optional[List[str]] = None,
        weights: Optional[Dict[str, Union[int, float]]] = None,
        annotations: Optional[Union[Annotations, Dict[str, Any]]] = None,
        uuid: Optional[Union[str, _uuid.UUID]] = None,
    ):
        self._uuid: _uuid.UUID
        self._name: str
        self._kind: str
        self._labels: List[str]
        self._weights: Dict[str, Union[int, float]]
        self._annotations: Annotations

        setattr(self, "uuid", uuid)
        setattr(self, "name", name)
        setattr(self, "kind", kind)
        setattr(self, "labels", labels)
        setattr(self, "weights", weights)
        setattr(self, "annotations", annotations)

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.uuid = None
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        result.uuid = None
        return result

    @property
    def uuid(self) -> _uuid.UUID:
        """Instance UUID."""
        return self._uuid

    @uuid.setter
    def uuid(self, value: Optional[Union[str, _uuid.UUID]]):
        if value is None:
            uuid = _uuid.uuid4()
        elif isinstance(value, _uuid.UUID):
            uuid = value
        else:
            uuid = _uuid.UUID(value)
        self._uuid = uuid

    @property
    def name(self) -> str:
        """Instance name. Given a UUID if None was provided."""
        return self._name

    @name.setter
    def name(self, value: Optional[str]):
        if value is None:
            value = str(self.uuid)
        self._name = str(value)

    @property
    def kind(self) -> str:
        """Kind or main category of this instance."""
        return self._kind

    @kind.setter
    def kind(self, value: str):
        if not value:
            self._kind = "default"
        else:
            self._kind = str(value)

    @property
    def labels(self) -> List[str]:
        """Labels categorizing this instance."""
        return self._labels

    @labels.setter
    def labels(self, value: List[str]):
        if not value:
            self._labels = ["default"]
        else:
            self._labels = list(value)

    @property
    def weights(self) -> Dict[str, Union[int, float]]:
        """Dictionary of weights attached to this instance."""
        return self._weights

    @weights.setter
    def weights(self, value: Optional[Dict[str, Union[int, float]]]):
        if not value:
            self._weights = dict(default=1)
        else:
            self._weights = dict(value)

    @property
    def weight(self) -> Union[int, float]:
        """Cumulative weight of this instance (read-only).

        Returns the sum of [`self.weights`][ragraph.generic.Metadata.weights].
        """
        return sum(self._weights.values())

    @property
    def annotations(self) -> Annotations:
        """Annotations of this instance.

        Defaults to an empty [`Annotations`][ragraph.generic.Annotations] instance.
        """
        return self._annotations

    @annotations.setter
    def annotations(self, value: Optional[Union[Annotations, Dict[str, Any]]]):
        if value is None:
            self._annotations = Annotations()
        elif isinstance(value, Annotations):
            self._annotations = value
        else:
            self._annotations = Annotations(**value)


class MetadataFilter:
    """Metadata filtering options.

    Arguments:
        uuids: Filter by UUIDs.
        names: Filter by names.
        kinds: Filter by kinds.
        labels: Filter by labels. Items should match at least one.
        weights: Filter by weight labels. Items should match at least one.
        weight_domains: Filter items by weight domains. (upper/lower bound)
        annotations: Filter by annotation keys. Items should match at least one.
    """

    _filters = [
        "uuids",
        "names",
        "kinds",
        "labels",
        "weights",
        "weight_domains",
        "annotations",
    ]

    def __init__(
        self,
        uuids: Optional[Union[Iterable[str], Iterable[_uuid.UUID]]] = None,
        names: Optional[Iterable[str]] = None,
        kinds: Optional[Iterable[str]] = None,
        labels: Optional[Iterable[str]] = None,
        weights: Optional[Iterable[str]] = None,
        weight_domains: Optional[Dict[str, ContinuousDomain]] = None,
        annotations: Optional[Iterable[str]] = None,
    ):
        if uuids:
            self.uuids: Set[_uuid.UUID] = {
                i if isinstance(i, _uuid.UUID) else _uuid.UUID(i) for i in uuids
            }
        if names:
            self.names: Set[str] = set(names)
        if kinds:
            self.kinds: Set[str] = set(kinds)
        if labels:
            self.labels: Set[str] = set(labels)
        if weights:
            self.weights: Set[str] = set(weights)
        if weight_domains:
            self.weight_domains: Dict[str, ContinuousDomain] = weight_domains
        if annotations:
            self.annotations: Set[str] = set(annotations)

    def __call__(self, item: Metadata) -> bool:
        """Apply this metadata filter to an item."""
        return all(check(item) for check in self.get_checks())

    def get_checks(self) -> List[Callable[[Metadata], bool]]:
        """Active filter methods."""
        return [
            getattr(self, f"check_{filter}")
            for filter in self._filters
            if getattr(self, filter, False)
        ]

    def filter(
        self, data: Iterable[Metadata], as_list: bool = True
    ) -> Union[List[Metadata], Generator[Metadata, None, None]]:
        """Filter data using the set metadata filters."""

        def _generator():
            checks = self.get_checks()
            if checks:
                for item in data:
                    if all(check(item) for check in checks):
                        yield item
            else:
                yield from data

        generator = _generator()
        return list(generator) if as_list else generator

    def check_uuids(self, item: Metadata) -> bool:
        """Check if item satisfies UUID filter."""
        return item.uuid in self.uuids

    def check_names(self, item: Metadata) -> bool:
        """Check if item satisfies names filter."""
        return item.name in self.names

    def check_kinds(self, item: Metadata) -> bool:
        """Check if item satisfies kinds filter."""
        return item.kind in self.kinds

    def check_labels(self, item: Metadata) -> bool:
        """Check if item satisfies labels filter."""
        return any(self.labels.intersection(item.labels))

    def check_weights(self, item: Metadata) -> bool:
        """Check if item satisfies weight keys filter."""
        return any(self.weights.intersection(item.weights.keys()))

    def check_weight_domains(self, item: Metadata) -> bool:
        """Check if item satisfies weight domains filter."""
        return all(
            (key in item.weights and item.weights[key] in domain)
            for key, domain in self.weight_domains.items()
        )

    def check_annotations(self, item: Metadata) -> bool:
        """Check if item satisfies annotation keys filter."""
        return any(self.annotations.intersection(item.annotations.keys()))


class MetadataOptions:
    """Seen values in an iterable of Metadata instances.

    Arguments:
        objects: Objects derivated of the Metadata class.
        skip_names: Whether to skip the names field.
    """

    def __init__(
        self,
        objects: Iterable[Metadata],
        skip_uuids: bool = False,
        skip_names: bool = False,
    ):
        uuids: Optional[Set[_uuid.UUID]] = None if skip_uuids else set()
        names: Optional[Set[str]] = None if skip_names else set()

        kinds: Set[str] = set()
        labels: Set[str] = set()
        weights: Set[str] = set()
        weight_domains: Dict[str, ContinuousDomain] = defaultdict(ContinuousDomain)
        annotations: Set[str] = set()

        for obj in objects:
            if uuids is not None:
                uuids.add(obj.uuid)
            if names is not None:
                names.add(obj.name)
            kinds.add(obj.kind)
            labels.update(obj.labels)
            weights.update(obj.weights.keys())
            annotations.update(obj.annotations.keys())

            for k, v in obj.weights.items():
                domain = weight_domains[k]
                if domain.lower is None or v < domain.lower:
                    domain.lower = Bound(value=v, inclusive=True, report=None)
                if domain.upper is None or v > domain.upper:
                    domain.upper = Bound(value=v, inclusive=True, report=None)

        self.uuids = uuids
        self.names = names
        self.kinds = kinds
        self.labels = labels
        self.weights = weights
        self.weight_domains = weight_domains
        self.annotations = annotations

    def as_dict(self) -> Dict[str, Any]:
        """Serializable dictionary representation."""
        result: Dict[str, Any] = dict(
            kinds=sorted(self.kinds),
            labels=sorted(self.labels),
            weights=sorted(self.weights),
            weight_domains={k: v.as_dict() for k, v in self.weight_domains.items()},
            annotations=sorted(self.annotations),
        )
        if self.uuids is not None:
            result["uuids"] = sorted(self.uuids, key=lambda x: str(x))
        if self.names is not None:
            result["names"] = sorted(self.names)

        return result



"""# JSON format support"""
import json
from pathlib import Path
from typing import Any, Dict, Optional, Union
from uuid import UUID

from ragraph.edge import Edge
from ragraph.generic import Annotations
from ragraph.graph import Graph
from ragraph.node import Node


class GraphEncoder(json.JSONEncoder):
    def default(self, obj):
        if (
            isinstance(obj, Graph)
            or isinstance(obj, Edge)
            or isinstance(obj, Node)
            or isinstance(obj, Annotations)
        ):
            return obj.json_dict
        if isinstance(obj, UUID):
            return str(obj)
        return json.JSONEncoder.encode(obj)


def graph_from_json_dict(graph_dict: Dict[str, Any], use_uuid: bool = True) -> Graph:
    """Recreate Graph object from JSON dictionary."""
    nodes = graph_dict.get("nodes", dict())
    edges = graph_dict.get("edges", dict())
    g = Graph(
        name=graph_dict.get("name"),
        kind=graph_dict.get("kind"),
        labels=graph_dict.get("labels"),
        weights=graph_dict.get("weights"),
        annotations=graph_dict.get("annotations"),
        uuid=graph_dict.get("uuid"),
    )

    # Get node dictionaries.
    if isinstance(nodes, dict):
        node_dicts = list(nodes.values())
    elif isinstance(nodes, list):
        node_dicts = nodes
    else:
        raise ValueError("Unrecognizable nodes property.")

    # Get edge dictionaries.
    if isinstance(edges, dict):
        edge_dicts = list(edges.values())
    elif isinstance(edges, list):
        edge_dicts = edges
    else:
        raise ValueError("Unrecognizable edges property.")

    # Add initial nodes.
    for nd in node_dicts:
        g.add_node(
            Node(
                name=nd.get("name"),
                kind=nd.get("kind"),
                labels=nd.get("labels"),
                parent=None,
                children=[],
                weights=nd.get("weights"),
                annotations=nd.get("annotations"),
                uuid=nd.get("uuid"),
            )
        )

    node_ref = {str(node.uuid): node for node in g.nodes} if use_uuid else g.node_dict
    ref_key = "uuid" if use_uuid else "name"

    # Resolve references.
    for nd in node_dicts:
        if nd.get("parent", False):
            node_ref[nd[ref_key]].parent = node_ref[nd["parent"]]
        if nd.get("children", False):
            node_ref[nd[ref_key]].children = [node_ref[c] for c in nd["children"]]
        if nd.get("is_bus", False):
            node_ref[nd[ref_key]].is_bus = True

    # Add edges.
    for ed in edge_dicts:
        g.add_edge(
            Edge(
                node_ref[ed["source"]],
                node_ref[ed["target"]],
                name=ed.get("name"),
                kind=ed.get("kind"),
                labels=ed.get("labels"),
                weights=ed.get("weights"),
                annotations=ed.get("annotations"),
                uuid=ed.get("uuid"),
            ),
        )

    return g


def from_json(
    path: Optional[Union[str, Path]] = None,
    enc: Optional[str] = None,
    use_uuid: bool = True,
) -> Graph:
    """Decode JSON file or string into a Graph.

    Arguments:
        path: JSON file path.
        enc: JSON encoded string.

    Returns:
        Graph object.
    """
    if path is None and enc is None:
        raise ValueError("`path` and `enc` arguments cannot both be `None`.")
    if path is not None and enc is not None:
        raise ValueError("`path` and `enc` arguments cannot both be set.")

    if path:
        enc = Path(path).read_text(encoding="utf-8")

    json_dict = json.loads(enc)

    return graph_from_json_dict(json_dict, use_uuid=use_uuid)


def to_json(graph: Graph, path: Optional[Union[str, Path]] = None) -> Optional[str]:
    """Encode Graph to JSON file or string.

    Arguments:
        path: Optional file path to write JSON to.

    Returns:
        JSON string.
    """

    enc = json.dumps(graph, cls=GraphEncoder, sort_keys=True, indent=4)
    if path is not None:
        path = Path(path)
        path.write_text(enc, encoding="utf-8")
        return None
    else:
        return enc



"""
# Sequencing algorithms

Sequencing algorithms search for sequences of nodes according to some objective or
logic. Simple are sorting by name and more elaborate examples sort according to Markov
transition probabilities or sequences without feedback (edges to earlier nodes).


## Available algorithms

- [`markov`][ragraph.analysis.sequence.markov] sequencing based on Markov transition probabilities
  between nodes.
- [`scc_tearing`][ragraph.analysis.sequence.scc_tearing] efficient algorithm to sort the "trivial"
  part of sequencing quickly by putting the Strongly Connected Components (largest possible cycles)
  in a non-feedback order and then "tears" iteratively according to a tearing indicator.
- [`name`][ragraph.analysis.sequence.name] just sorts nodes by their name.
- [`tarjans_dfs`][ragraph.analysis.sequence.tarjans_dfs] Tarjan's Depth First Search algorithm.
  Efficient sorting algorithm for Directed Acyclic Graphs (DAG).
- [`axis`][ragraph.analysis.sequence.axis] Typical Multi-Domain Matrix axis sorting method.


## Metrics

There are several metrics from literature available to grade sequences as well. The
metrics are documented over at [`metrics`][ragraph.analysis.sequence.metrics].


# Utilities

Finally, there are some utilities like branch-sorting (recursively sorting all branches
in a hierarchical tree instead of all leaf nodes as a whole) available in
[`utils`][ragraph.analysis.sequence.utils]:

- [`branchsort`][ragraph.analysis.sequence.branchsort] Recursively sort children within a branch or
  hierarchy using any of the above.
"""

from ragraph.analysis.sequence import metrics, utils
from ragraph.analysis.sequence._axis import axis
from ragraph.analysis.sequence._genetic import genetic
from ragraph.analysis.sequence._markov import markov
from ragraph.analysis.sequence._name import name
from ragraph.analysis.sequence._scc_tearing import scc_tearing
from ragraph.analysis.sequence._tarjan import tarjans_dfs
from ragraph.analysis.sequence.utils import branchsort

__all__ = [
    "metrics",
    "utils",
    "axis",
    "genetic",
    "markov",
    "name",
    "scc_tearing",
    "tarjans_dfs",
    "branchsort",
]



"""# Cubehelix color palette

Reference:
    Green, D. A. (2011). A colour scheme for the display of astronomical intensity images.
    Bull. Astr. Soc. India. Retrieved from
    [http://www.astro.caltech.edu/](http://www.astro.caltech.edu/).
"""

from typing import List, Optional

from ragraph.colors import utils

try:
    import numpy as np

    HAVE_NUMPY = True
except ImportError:
    HAVE_NUMPY = False


def cubehelix_palette(
    n_colors: int,
    start: float = 2.25,
    end: float = 3.45,
    dark: float = 0.2,
    light: float = 0.85,
    gamma: float = 1.0,
    saturation: float = 1.2,
    rotations: Optional[float] = None,
    categorical: bool = False,
) -> List[str]:
    """Calculate a cubehelix color palette as a list of hex codes.

    Arguments:
        n_colors: Number of RGB colour pairs to generate.
        start: Starting color (Red: 1.0, Green: 2.0, Blue: 3.0).
        end: Final color (Red: 1.0, Green: 2.0, Blue: 3.0).
        dark: Lightness of the darkest value [0.0 - 1.0].
        light: Lightness of the lightest value [0.0 - 1.0].
        gamma: Exponential factor applied to lightness to emphasize low intensity values (< 1.0) or
            high intensity values (> 1.0).
        rotations: Number of RGB rotations (1.0 means a full RGB rotation). Overrides `end`.
            Otherwise calculated as `(end - start) / 3`.
        categorical: Whether to shuffle the rotational space optimally for categorical color
            palettes. Similar colors are positioned as far apart as possible.

    Returns:
        Color palette as a list of hex codes.

    Reference:
        Green, D. A. (2011). A colour scheme for the display of astronomical intensity images. Bull.
        Astr. Soc. India. Retrieved from http://www.astro.caltech.edu/
    """
    if HAVE_NUMPY:
        method = _cubehelix_numpy
    else:
        method = _cubehelix_pure_python

    n_colors = max(n_colors, 1)

    rgb1s = method(
        n_colors,
        start=start,
        end=end,
        dark=dark,
        light=light,
        gamma=gamma,
        saturation=saturation,
        rotations=rotations,
        categorical=categorical,
    )

    hexes = [utils.rgb_to_hex(*utils.rgb1_to_rgb(*c)) for c in rgb1s]

    return hexes


def _cubehelix_numpy(
    n_colors: int,
    start: float = 2.25,
    end: float = 3.45,
    dark: float = 0.2,
    light: float = 0.85,
    gamma: float = 1.0,
    saturation: float = 1.2,
    rotations: Optional[float] = None,
    categorical: bool = False,
) -> List[List[float]]:
    """See [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette]."""
    n_colors = max(n_colors, 1)

    if rotations is None:
        rotations = (end - start) / 3

    start = start / 3

    light_space = np.linspace(light, dark, n_colors, endpoint=True) ** gamma
    a_space = saturation / 2 * light_space * (1 - light_space)

    phi_space = 2 * np.pi * np.linspace(start, start + rotations, n_colors)
    if categorical:
        phi_space = np.array(_shuffle(phi_space, n_colors // 3))
    cos_space = np.cos(phi_space)
    sin_space = np.sin(phi_space)

    rot_mat = np.vstack((cos_space, sin_space))

    # Constants from paper.
    col_mat = np.array([[-0.14861, 1.78277], [-0.29227, -0.90649], [1.97294, 0.0]])

    colors = light_space + a_space * (col_mat @ rot_mat)
    colors[colors > 1.0] = 1.0
    colors[colors < 0.0] = 0.0

    return colors.T.tolist()


def _cubehelix_pure_python(
    n_colors: int,
    start: float = 2.25,
    end: float = 3.45,
    dark: float = 0.2,
    light: float = 0.85,
    gamma: float = 1.0,
    saturation: float = 1.2,
    rotations: Optional[float] = None,
    categorical: bool = False,
) -> List[List[float]]:
    """See [`cubehelix_palette`][ragraph.colors.cubehelix.cubehelix_palette]."""
    import math

    if rotations is None:
        rotations = (end - start) / 3

    start = start / 3

    light_space = [math.pow(x, gamma) for x in _linspace(light, dark, n_colors)]
    a_space = [saturation / 2 * x * (1 - x) for x in light_space]

    phi_space = [2 * math.pi * x for x in _linspace(start, start + rotations, n_colors)]
    if categorical:
        phi_space = _shuffle(phi_space, n_colors // 3)
    cos_space = [math.cos(x) for x in phi_space]
    sin_space = [math.sin(x) for x in phi_space]

    col_mat = [[-0.14861, 1.78277], [-0.29227, -0.90649], [1.97294, 0.0]]

    colors = []
    for light, a, cos, sin in zip(light_space, a_space, cos_space, sin_space):
        color = []
        for col_row in col_mat:
            value = light + a * (col_row[0] * cos + col_row[1] * sin)
            value = min(1.0, max(0.0, value))
            color.append(value)
        colors.append(color)
    return colors


def _linspace(lower: float, upper: float, n: int) -> List[float]:
    """Linspace between lower and upper bound (inclusive) without Numpy."""
    step = (upper - lower) / (n - 1)
    return [lower + i * step for i in range(n)]


def _shuffle(items: List, n_splits: int = 2) -> List:
    """Shuffle items from by iterating between splits of the items."""
    if n_splits <= 1:
        return items

    n_items = len(items)
    frac = n_items / n_splits
    if frac <= 1:
        return items

    shuffled = [None] * n_items
    taken = 0
    for i in range(n_splits):
        take = len(shuffled[i::n_splits])
        shuffled[i::n_splits] = items[taken : taken + take]
        taken += take

    return shuffled



"""# Node class module"""

from typing import Any, Dict, Generator, Iterable, List, Optional, Union
from uuid import UUID

from ragraph.generic import Annotations, Metadata


class Node(Metadata):
    """Generic node class.

    Arguments:
        name: Instance name. Given a UUID if none provided.
        parent: Parent node.
        children: List of child nodes.
        kind: Kind or main category of this instance.
        labels: Labels categorizing this instance.
        weights: Dictionary of weights attached to this instance.
        annotations: Miscellaneous properties of this instance.
        uuid: Fixed UUID if desired, generated when left set to None.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        parent: Optional["Node"] = None,
        children: Optional[Iterable["Node"]] = None,
        is_bus: bool = False,
        kind: str = "node",
        labels: Optional[List[str]] = None,
        weights: Optional[Dict[str, Union[int, float]]] = None,
        annotations: Union[Annotations, Dict[str, Any], None] = None,
        uuid: Optional[Union[str, UUID]] = None,
    ):
        Metadata.__init__(
            self,
            name=name,
            kind=kind,
            labels=labels,
            weights=weights,
            annotations=annotations,
            uuid=uuid,
        )
        self._parent: Optional[Node] = None
        self._children: List[Node] = []
        self._is_bus: bool = False
        setattr(self, "parent", parent)
        setattr(self, "children", children)
        setattr(self, "is_bus", is_bus)

    def __str__(self) -> str:
        return "{}(name='{}')".format(self.__class__.__name__, self.name)

    def __repr__(self) -> str:
        parent = None if self.parent is None else repr(self.parent.name)
        kwargdict = dict(
            name=repr(self.name),
            parent=parent,
            children=[str(c) for c in self.children],
            is_bus=self.is_bus,
            kind=repr(self.kind),
            labels=self.labels,
            weights=self.weights,
            annotations=self.annotations,
            uuid=repr(self.uuid),
        )
        kwargs = ", ".join([str(k) + "=" + str(v) for k, v in kwargdict.items()])
        return "<{}.{}({}) at {}>".format(
            self.__class__.__module__, self.__class__.__name__, kwargs, hex(id(self))
        )

    @property
    def parent(self) -> Optional["Node"]:
        """Parent node. Defaults to `None`."""
        return self._parent

    @parent.setter
    def parent(self, value: Optional["Node"]):
        # Check type.
        if value is not None and not isinstance(value, Node):
            raise TypeError("Parent should be of type {}, got {}.".format(Node, type(value)))
        previous = getattr(self, "_parent")
        if previous is not None:
            # Previous parent, need to unset first!
            previous._children.remove(self)
            self._parent = None
        if value is not None:
            # Re-set if value is not None.
            self._parent = value
            self._parent._children.append(self)

    @property
    def children(self) -> List["Node"]:
        """Child nodes. Default is empty."""
        return self._children

    @children.setter
    def children(self, value: Optional[Iterable["Node"]]):
        # Unset previous children's private parent properties, bypassing parent setter.
        for c in self._children:
            c._parent = None
        self._children = []
        # New children: set parent publicly to also unset any previous parent and adds
        # it to self._children.
        if value is not None:
            for c in list(value).copy():
                c.parent = self

    @property
    def is_bus(self):
        """Whether this node is a bus node for its parent."""
        return self._is_bus

    @is_bus.setter
    def is_bus(self, value: bool):
        value = bool(value)
        if value and not self._parent:
            raise ValueError("Cannot be a bus without a parent to be it for. Set a parent first!")
        self._is_bus = bool(value)

    @property
    def ancestors(self) -> List["Node"]:
        """Get all ancestors of this node."""
        return list(self.ancestor_gen)

    @property
    def ancestor_gen(self) -> Generator["Node", None, None]:
        """Yield all ancestors of this node."""
        if self.parent:
            yield self.parent
            yield from self.parent.ancestors

    @property
    def descendants(self) -> List["Node"]:
        """Get all descendants of this node."""
        return list(self.descendant_gen)

    @property
    def descendant_gen(self) -> Generator["Node", None, None]:
        """Yield all descendants of this node."""
        for child in self.children:
            yield child
            yield from child.descendants

    @property
    def depth(self) -> int:
        """Depth of this node, e.g. the number of levels from the node to the root."""
        if self.parent:
            return self.parent.depth + 1
        return 0

    @property
    def height(self) -> int:
        """Height of this node, e.g. the longest simple path to a leaf node."""
        if self._children:
            return max([child.height for child in self._children]) + 1
        return 0

    @property
    def width(self) -> int:
        """Width of this node, e.g. the number of leaf nodes in this node's branch."""
        if self.is_leaf:
            return 1
        return sum(c.width for c in self.children)

    @property
    def is_leaf(self) -> bool:
        """Whether this node is a leaf node (has no children)."""
        return not bool(self.children)

    @property
    def is_root(self) -> bool:
        """Whether this node is a root node (has no parent)."""
        return not bool(self.parent)

    @property
    def json_dict(self) -> Dict[str, Any]:
        """JSON dictionary representation.

        Returns:
            name: Name as str.
            parent: Parent name (not Node) as str.
            children: Children names (not Nodes) as str.
            is_bus: is_bus as bool.
            kind: Kind as str.
            labels: Labels as a list of strings.
            weights: Weights as dict.
            annotations: Annotations as per
                [`ragraph.generic.Mapping.as_dict`][ragraph.generic.Mapping.as_dict].
        """
        return self.as_dict(use_uuid=True)

    def as_dict(self, use_uuid: bool = False) -> Dict[str, Any]:
        """Return a copy as a (serializable) dictionary.

        Returns:
            name: Name as str.
            parent: Parent name or UUID (not Node) as str.
            children: Children names (not Nodes) as str.
            is_bus: is_bus as bool.
            kind: Kind as str.
            labels: Labels as a list of strings.
            weights: Weights as dict.
            annotations: Annotations as per
                [`ragraph.generic.Mapping.as_dict`][ragraph.generic.Mapping.as_dict].
            uuid: UUID as str if toggled.
        """
        if use_uuid:
            return dict(
                name=self.name,
                parent=None if self.parent is None else str(self.parent.uuid),
                children=[str(n.uuid) for n in self.children],
                is_bus=self.is_bus,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
                uuid=str(self.uuid),
            )
        else:
            return dict(
                name=self.name,
                parent=getattr(self.parent, "name", None),
                children=[n.name for n in self.children],
                is_bus=self.is_bus,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
            )



"""# SVG shapes

This module contains two base classes, being the [`Line`][ragraph.plot.svg.Line] options for SVG
shapes and an [`SVG`][ragraph.plot.svg.SVG] mapping for the shapes themselves. You will also find
several methods to obtain several basic shapes conveniently.
"""
from math import cos, pi, sin
from typing import Any, Dict, Optional, Union

from ragraph.generic import Mapping, field


class Line(Mapping):
    """SVG line options mapping.

    Arguments:
        color: Line color.
        width: Line width in pixels.
        dash: Dash style of the line.
    """

    _defaults = dict(color="rgba(0,0,0,0)", width=2, dash="solid")

    def __init__(
        self,
        color: Optional[str] = None,
        width: Optional[int] = None,
        dash: Optional[str] = None,
    ):
        super().__init__(color=color, width=width, dash=dash)

    @field
    def color(self) -> str:  # type: ignore
        """Line color."""

    @field
    def width(self) -> int:  # type: ignore
        """Line width in pixels."""

    @field
    def dash(self) -> str:  # type: ignore
        """Dash style of the line."""


class SVG(Mapping):
    """SVG shape mapping.

    Arguments:
        visible: Toggles shape visibility.
        type: One of 'circle', 'rect', 'path' or 'line'.
        layer: 'above' draws shape above traces, 'below' under them.
        xref: x coordinate axis. 'paper' or 'x', 'x1', 'x2', etc.
        xsizemode: 'scaled' or 'pixel'. Relative or absolute sizing w.r.t. axis.
        xanchor: If sizemode is set to 'pixel', reference on axis to anchor shape to.
        x0: Starting x position.
        x1: Ending x position.
        yref: y coordinate axis. 'paper' or 'y', 'y1', 'y2', etc.
        ysizemode: 'scaled' or 'pixel'. Relative or absolute sizing w.r.t. axis.
        yanchor: If sizemode is set to 'pixel', reference on axis to anchor shape to.
        y0: Starting y position.
        y1: Ending y position.
        path: For shapetype 'path', a valid SVG path, with data values as coordinates when
            referencing axis and absolute pixels with respect to anchors when the 'pixel' sizemode
            is set.
        opacity: The opacity between 0.0 and 1.0.
        line: Line mapping options. See [`Line`][ragraph.plot.svg.Line].
        fillcolor: Interior shape color.
        fillrule: Determines which regions of complex paths constitute the interior.
            One of "evenodd" or "nonzero".
        editable: Whether the shape could be activated for edit or not.
        name: Only used with templates.
        templateitemname: Used to refer to a named item in this array in the template.
    """

    _keys = {
        "visible",
        "type",
        "layer",
        "xref",
        "xsizemode",
        "xanchor",
        "x0",
        "x1",
        "yref",
        "ysizemode",
        "yanchor",
        "y0",
        "y1",
        "path",
        "opacity",
        "line",
        "fillcolor",
        "fillrule",
        "editable",
        "name",
        "templateitemname",
    }

    def __init__(
        self,
        visible: Optional[bool] = None,
        type: Optional[str] = None,
        layer: Optional[str] = None,
        xref: Optional[str] = None,
        xsizemode: Optional[str] = None,
        xanchor: Optional[str] = None,
        x0: Optional[float] = None,
        x1: Optional[float] = None,
        yref: Optional[str] = None,
        ysizemode: Optional[str] = None,
        yanchor: Optional[str] = None,
        y0: Optional[float] = None,
        y1: Optional[float] = None,
        path: Optional[str] = None,
        opacity: Optional[float] = None,
        line: Optional[Union[Dict[str, Any], Line]] = None,
        fillcolor: Optional[str] = None,
        fillrule: Optional[str] = None,
        editable: Optional[bool] = None,
        name: Optional[str] = None,
        templateitemname: Optional[str] = None,
    ):
        super().__init__(
            visible=visible,
            type=type,
            layer=layer,
            xref=xref,
            xsizemode=xsizemode,
            xanchor=xanchor,
            x0=x0,
            x1=x1,
            yref=yref,
            ysizemode=ysizemode,
            yanchor=yanchor,
            y0=y0,
            y1=y1,
            path=path,
            opacity=opacity,
            line=line,
            fillcolor=fillcolor,
            fillrule=fillrule,
            editable=editable,
            name=name,
            templateitemname=templateitemname,
        )

    @field
    def visible(self) -> bool:  # type: ignore
        """Toggles shape visibility."""

    @field
    def type(self) -> str:  # type: ignore
        """One of 'circle', 'rect', 'path' or 'line'."""

    @field
    def layer(self) -> str:  # type: ignore
        """'above' draws shape above traces, 'below' under them."""

    @field
    def xref(self) -> str:  # type: ignore
        """x coordinate axis. 'paper' or 'x', 'x1', 'x2', etc."""

    @field
    def xsizemode(self) -> str:  # type: ignore
        """'scaled' or 'pixel'. Relative or absolute sizing w.r.t. axis."""

    @field
    def xanchor(self) -> str:  # type: ignore
        """If sizemode is set to 'pixel', reference on axis to anchor shape to."""

    @field
    def x0(self) -> float:  # type: ignore
        """Starting x position."""

    @field
    def x1(self) -> float:  # type: ignore
        """Ending x position."""

    @field
    def yref(self) -> str:  # type: ignore
        """y coordinate axis. 'paper' or 'y', 'y1', 'y2', etc."""

    @field
    def ysizemode(self) -> str:  # type: ignore
        """'scaled' or 'pixel'. Relative or absolute sizing w.r.t. axis."""

    @field
    def yanchor(self) -> str:  # type: ignore
        """If sizemode is set to 'pixel', reference on axis to anchor shape to."""

    @field
    def y0(self) -> float:  # type: ignore
        """Starting y position."""

    @field
    def y1(self) -> float:  # type: ignore
        """Ending y position."""

    @field
    def path(self) -> str:  # type: ignore
        """For shapetype 'path', a valid SVG path, with data values as coordinates when referencing
        axis and absolute pixels with respect to anchors when the 'pixel' sizemode is set.
        """

    @field
    def opacity(self) -> float:  # type: ignore
        """The opacity between 0.0 and 1.0."""

    @field
    def line(self) -> Line:  # type: ignore
        """Line mapping options. See [`Line`][ragraph.plot.svg.Line]."""

    @field
    def fillcolor(self) -> str:  # type: ignore
        """Interior shape color."""

    @field
    def fillrule(self) -> str:  # type: ignore
        """Determines which regions of complex paths constitute the interior.
        One of 'evenodd' or 'nonzero'.
        """

    @field
    def editable(self) -> bool:  # type: ignore
        """Whether the shape could be activated for edit or not."""

    @field
    def name(self) -> str:  # type: ignore
        """Only used with templates."""

    @field
    def templateitemname(self) -> str:  # type: ignore
        """Used to refer to a named item in this array in the template."""


def get_line(
    x0: float,
    x1: float,
    y0: float,
    y1: float,
    **kwargs: Any,
) -> SVG:
    """Get straight line SVG mapping.

    Arguments:
        x0: Starting x position.
        x1: Ending x position.
        y0: Starting y position.
        y1: Ending y position.
        **kwargs: Overrides for the [`SVG`][ragraph.plot.svg.SVG] object.

    Returns:
        SVG shape mapping.
    """
    return SVG(type="line", x0=x0, x1=x1, y0=y0, y1=y1, **kwargs)


def get_curvedline(
    x0: float,
    x1: float,
    x2: float,
    y0: float,
    y1: float,
    y2: float,
    **kwargs: Any,
) -> SVG:
    """Get curved line (quadratic Bezier) SVG mapping.

    Arguments:
        x0: Starting x position.
        x1: Control point x position.
        x2: Ending x position.
        y0: Starting y position.
        y1: Control point y position.
        y2: Ending y position.
        **kwargs: Overrides for the [`SVG`][ragraph.plot.svg.SVG] object.

    Returns:
        Quadratic Bezier SVG shape mapping.
    """
    path = f"M {x0} {y0} Q {x1} {y1} {x2} {y2}"
    return SVG(type="path", path=path, **kwargs)


def get_rectangle(
    x0: float,
    x1: float,
    y0: float,
    y1: float,
    **kwargs: Any,
) -> SVG:
    """Get a rectangle SVG mapping.

    Arguments:
        x0: Starting x position.
        x1: Ending x position.
        y0: Starting y position.
        y1: Ending y position.
        **kwargs: Overrides for the [`SVG`][ragraph.plot.svg.SVG] object.

    Returns:
        SVG shape mapping.
    """
    return SVG(type="rect", x0=x0, x1=x1, y0=y0, y1=y1, **kwargs)


def get_wedge(
    x: float, y: float, r: float, start_angle: float, end_angle: float, **kwargs: Any
) -> SVG:
    """Get a wedge SVG mapping.

    Arguments:
        x: x position of the wedge center.
        y: y position of the wedge center.
        r: radius of the wedge.
        start_angle: Starting angle (radians) of the wedge.
        end_angle: Ending angle (radians) of the wedge.
        **kwargs: Overrides for the [`SVG`][ragraph.plot.svg.SVG] object.

    Returns:
        SVG shape mapping.
    """
    segments = ["M {} {}".format(x, y)]
    xa = x + round(r * cos(start_angle), 3)
    xb = y + round(r * sin(start_angle), 3)
    segments.append(" L {} {}".format(xa, xb))

    delta = end_angle - start_angle
    max_delta = 0.25 * pi
    while delta > 0:
        if delta >= max_delta:
            angle_step = max_delta
        else:
            angle_step = delta

        # Get curve control points
        xc = round(
            x + r / cos(angle_step / 2) * cos(start_angle + angle_step / 2),
            3,
        )
        yc = round(
            y + r / cos(angle_step / 2) * sin(start_angle + angle_step / 2),
            3,
        )
        xe = round(x + r * cos(start_angle + angle_step), 3)
        ye = round(y + r * sin(start_angle + angle_step), 3)

        # Add the segment.
        segments.append(" Q {} {} {} {}".format(xc, yc, xe, ye))
        delta -= angle_step
        start_angle += angle_step

    segments.append(" Z")

    path = "".join(segments)

    return SVG(type="path", path=path, **kwargs)



"""Graph of a design problem complexity study. Components are related (have edges) to
attributes that are of importance, which are of edge kind "incidence". Nodes have kind
"component" or "attribute".

Reference:
    [Figure 1(a)] Chen, L., & Li, S. (2005). Analysis of Decomposability and
    Complexity for Design Problems in the Context of Decomposition. Journal of Mechanical
    Design, 127(4), 545. [DOI: 10.1115/1.1897405](https://doi.org/10.1115/1.1897405)
"""



"""Ford Motor Company Hood Development Process.

This DSM application was made to test the effectiveness of a DSM analysis to improve a highly
developed iterative design process in terms of:

- Reduction of product development lead time
- Reduction of product development lead-time variation

As it is a process DSM, the nodes represent tasks in the design process. The edges represent
dependencies between them. Each task in the design process was assigned a task volatility value,
which is an indication of the probability of rework.

Node (task) weights:

- EC(i) and EC(r) represent the initial and rework costs
- ED(i) and ED(r) represent the initial and rework durations
- Information variability (likelihood of input changing)


Edge weights:

- volatility: an indication of the probability of rework when the dependent task is executed without
  properly waiting for its input to finish. It's value was determined using the information
  variability of the source (depended on) task and the sensitivity of the target (dependent) task.

Reference:
    Browning, Tyson R., and Steven D. Eppinger. 2002, November. Modeling Impacts of Process
    Architecture on Cost and Schedule Risk in Product Development. IEEE Transactions on Engineering
    Management 49 (4):428-442.

Reference:
    Krishnan, Viswanathan, Steven D. Eppinger, and Daniel E. Whitney. 1997, April. A Model-Based
    Framework to Overlap Product Development Activities. Management Science 43 (4):437-451.

Reference:
    Yassine, Ali A., Daniel E. Whitney, and Tony P. Zambito. 2001. Assessment of Rework
    Probabilities for Simulating Product Development Processes Using the Design Structure Matrix
    (DSM). ASME Design Engineering Technical Conferences, DTM-21693.
"""


node_weights = [
    "EC(i) ($000)",
    "EC(r) ($000)",
    "ED(i) (days)",
    "ED(r) (% of ED (i))",
    "information variability",
]
edge_weights = ["volatility"]



"""Genetic algorithms for sequencing purposes."""
from typing import List, Optional, Tuple, Union

from ratio_genetic_py import (
    ConvergenceKinds,
    CrossoverKinds,
    EvaluatorKinds,
    EvaluatorMatrix,
    GeneratorKinds,
    Lineage,
    MutatorSwap,
    RecorderKinds,
    SelectorKinds,
    SequencingSettings,
    sequence_sga,
)

from ragraph.analysis._classes import Parameter, SequenceAnalysis
from ragraph.graph import Graph, Node

params = {
    p.name: p
    for p in [
        Parameter(
            "n_chromosomes",
            int,
            description="Number of chromosomes in a generation.",
            default=100,
        ),
        Parameter(
            "n_generations",
            int,
            description="How many generations to go through after which to exit.",
            default=1000,
        ),
        Parameter(
            "p_crossover",
            float,
            description="Probability for a pair to be subjected to crossover.",
            default=0.3,
        ),
        Parameter(
            "p_mutation",
            float,
            description="Probability for each chromosome to be subjected to mutation.",
            default=0.05,
        ),
        Parameter(
            "p_swap",
            float,
            description="Probability for each gene to be swapped during mutation.",
            default=0.05,
        ),
        Parameter("evaluator", str, description="", default="feedback_distance"),
    ]
}


genetic_analysis = SequenceAnalysis("Genetic", parameters=params)


@genetic_analysis
def genetic(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[str], List[Node]]] = None,
    evaluator: Optional[str] = params["evaluator"].default,
    n_chromosomes: Optional[int] = params["n_chromosomes"].default,
    n_generations: Optional[int] = params["n_generations"].default,
    p_crossover: Optional[float] = params["p_crossover"].default,
    p_mutation: Optional[float] = params["p_mutation"].default,
    p_swap: Optional[float] = params["p_swap"].default,
    inherit: bool = True,
    edge_weights: Optional[List[str]] = None,
    loops: bool = False,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, List[Node]]:
    """docstring stub"""
    lineage = genetic_sequencing(
        graph,
        nodes,  # type: ignore
        n_chromosomes,  # type: ignore
        n_generations,  # type: ignore
        1,
        p_crossover,  # type: ignore
        p_mutation,  # type: ignore
        p_swap,  # type: ignore
        n_records=1,
        inherit=inherit,
        edge_weights=edge_weights,
    )

    best = lineage.hall_of_fame.chromosomes[0].genes

    seq = [nodes[i] for i in best]  # type: ignore
    return graph, seq


def genetic_sequencing(
    graph: Graph,
    nodes: List[Node],
    n_chromosomes: int,
    n_generations: int,
    n_hall_of_fame: int,
    p_crossover: float,
    p_mutation: float,
    p_swap: float,
    evaluator: str = "feedback_distance",
    n_records: Optional[int] = None,
    inherit: bool = True,
    edge_weights: Optional[List[str]] = None,
) -> Lineage:
    """Genetic sequencing of nodes in a graph.

    Arguments:
        graph: Graph holding data.
        nodes: Nodes to sequence.
        evaluator: Evaluation method to use. One of "feedback_distance",
            "feedback_marks", or "lower_left_distance".
        n_chromosomes: Number of chromosomes in each generation.
        n_generations: Number of generations to simulate.
        n_hall_of_fame: Hall of Fame size of best performing chromosomes.
        p_crossover: Probability for a pair to be subjected to crossover.
        p_mutation: Probability for each chromosome to be subjected to mutation.
        p_swap: Probability for each gene to be swapped with another during mutation.
        n_records: Number of generation records to keep.
        inherit: Whether to inherit edges between children when getting the adjacency
            matrix.
        edge_weights: Edge weights to consider when getting the adjacency matrix.

    Returns:
        Lineage object containing generations of chromosomes, generation records and
        a hall of fame of best performing chromosomes.
    """
    matrix = graph.get_adjacency_matrix(nodes, inherit=inherit, only=edge_weights)
    vec = [x for row in matrix for x in row]

    ekinds = dict(
        feedback_distance=EvaluatorKinds.FeedbackDistance,
        feedback_marks=EvaluatorKinds.FeedbackMarks,
        lower_left_distance=EvaluatorKinds.LowerLeftDistance,
    )
    evaluator = EvaluatorMatrix(ekinds[evaluator], vec, 1)

    settings = SequencingSettings(
        n_genes=len(nodes),
        p_crossover=p_crossover,
        p_mutation=p_mutation,
        n_chromosomes=n_chromosomes,
        n_generations=n_generations,
        n_records=n_records,
        n_hall_of_fame=n_hall_of_fame,
    )

    lineage = sequence_sga(
        settings,
        GeneratorKinds.RandomSequence,
        evaluator,
        RecorderKinds.FitnessStatistics,
        SelectorKinds.Roulette,
        CrossoverKinds.IPX,
        MutatorSwap(p_swap),
        ConvergenceKinds.Never,
    )

    return lineage



"""# Similarity analysis example

Example graph for similarity analysis. Products are related (have edges) to attributes that are of
importance in product portfolio analysis. Edges have edge kind "incidence". Nodes have kind
"product" or "attribute".
"""



"""# 8 node directed graph example from Robert Tarjan.

Reference:
    [Figure 3] Tarjan, R. (1972). Depth-First Search and Linear Graph Algorithms.
    SIAM Journal on Computing, 1(2), 146-160.
    [DOI: 10.1136/0201010](https://doi.org/10.1137/0201010)
"""

edge_weights = ["adjacency"]



"""# UCAV (unmanned combat aerial vehicle) preliminary design process.

The graph contains 14 preliminary design activities. Each phase consists of an initial
activity to define design several disciplines-such as aerodynamics, propulsion,
stability and control (S&C), distribute a design configuration (a design concept
proposed to satisfy the DR&O). Then, requirements and objectives (DR&O), followed by a
couple of activities to create and mechanical and electrical, weights, and performance -
each evaluate the configuration from their own perspective.

The nodes contain a minimum, mean and max weight for both duration and cost, as well as
an improvement curve value. The improvement curve determines the savings in work for
each successive iteration of an activity. The edges contain a binary weight and rework
probability and rework impact annotations.

When an activity is performed before all activities with edges to that activity has been
succesfully executed, there is a rework probability that the activity needs to be
redone, the impact then indicates the percentage that needs to be redone and the
improvement curve decreases this in successive iterations.

Reference:
    Eppinger, S. D., & Browning, T. R. (2012). Design Structure Matrix - Methods and Applications.
"""

node_weights = [
    "min_duration",
    "mean_duration",
    "max_duration",
    "min_cost",
    "mean_cost",
    "max_cost",
    "improvement_curve",
]
edge_weights = ["binary", "probability", "impact"]



"""# Pratt & Whitney Aircraft Engine

A directed graph describing the Pratt & Whitney PW4098 commercial high bypass-ratio turbofan engine.
The graph describes a combination of the actual hardware dependencies of the engine and those of the
development teams involved with them. It is a weighted and directed graph featuring 60 elements,
four of which are sometimes left out as they represent the integration teams and no individual
hardware components. Weak and strong dependencies are distinguished using weights of 1 and 2,
respectively.

Reference:
    Rowles, C. M. (1999). System integration analysis of a large commercial aircraft engine.
"""

edge_weights = ["adjacency"]



"""# Modular architecture example

Graph describing a design problem using objects that need to be designed based on certain aspects.
It is part of a set of three design problem graphs, each ideally solved with a different approach --
being a **modular**, integral or mixed architecture.

This particular graph describes a design problem which is ideally solved using a **modular object
architecture**. That means that the objects are best put into modules. Modules on different aspect
domains can then be mixed and matched so achieve a solution for it's combined members. Each module
should be designed based on all aspects that the objects members posess combined, while taking into
account the shared dependencies with other domains. A module's design solution should be applicable
to every member.

Nodes are of kind "object" or "aspect_1", "aspect_2", and "aspect_3". The numbers represent
different aspect domains. An "incidence" kind edge from an object to an aspect means that it
posesses that aspect.

Authors:
    Tim Wilschut and Tiemen Schuijbroek, October 2019.
"""



"""A sequencing algorithm for cyclic graphs that tears cycles based on a metric
(penalty) function. By default we use the metric that the Markov sequencing algorithm
uses to sort all nodes straight away.
"""
from textwrap import dedent
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

from ragraph.analysis._classes import MethodCast, Parameter, SequenceAnalysis
from ragraph.analysis.cluster._tarjan import tarjans_scc_algorithm
from ragraph.analysis.sequence.utils import markov_decision
from ragraph.graph import Graph
from ragraph.node import Node

params = {
    p.name: p
    for p in [
        Parameter(
            "inherit",
            bool,
            description="Whether to take into account (inherit) edges between "
            + "children during SCC calculations.",
            default=True,
        ),
        Parameter(
            "decision",
            Callable[
                [Graph, Union[List[Node], List[str]], Union[List[Node], List[str]]],
                int,
            ],
            description=(
                "Decision function to use. Should take a graph and a list of options ",
                "(nodes) and return an index of the selected option.",
            ),
            cast=MethodCast(),
            default=markov_decision,
        ),
        Parameter(
            "decision_args",
            Optional[Dict[str, Any]],
            description="Parameters to pass to the selected sequencing algorithm.",
            cast=lambda value: dict() if value is None else dict(value),
        ),
    ]
}


scc_tearing_analysis = SequenceAnalysis(
    "Strongly Connected Component Tearing",
    description=dedent(
        """\
    Sequencing by means of tearing Strongly Connected Components. First, the largest
    possible cycles are found and automatically sequenced using Tarjan's SCC algorithm.
    From each cycle, a node with the least penalty is then torn from the cycle and added
    to the sequence. The remainder is then again subjected to this procedure until all
    nodes have been sequenced.
    """
    ),
    parameters=params,
)


@scc_tearing_analysis
def scc_tearing(
    graph: Graph,
    decision: Callable[[Graph, List[Node]], Node],
    decision_args: Optional[Dict[str, Any]] = None,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, List[Node]]:
    """docstring stub"""
    sequence = scc_tearing_algorithm(
        graph, nodes, inherit, loops, decision, decision_args  # type: ignore
    )
    seq = list(sequence)

    return graph, seq


def scc_tearing_algorithm(
    graph: Graph,
    nodes: List[Node],
    inherit: bool,
    loops: bool,
    decision: Callable[[Graph, List[Node]], int],
    decision_args: Optional[Dict[str, Any]] = None,
) -> Generator[Node, None, None]:
    # Reverse to obtain topological order.
    sccs = tarjans_scc_algorithm(graph, nodes, inherit)[::-1]

    decision_args = dict() if decision_args is None else decision_args

    # Sequence to fill.
    for scc in sccs:
        # Sole node, add to sequence.
        if len(scc) == 1:
            sole = scc[0]
            scc_tearing_analysis.log(f"sole node '{sole.name}'")
            yield sole
            continue

        # Pick node to tear.
        tear = decision(graph, scc, **decision_args)
        torn = scc.pop(tear)
        scc_tearing_analysis.log(f"torn '{torn.name}' from {[n.name for n in scc]}.")
        yield torn
        yield from scc_tearing_algorithm(graph, scc, inherit, loops, decision, decision_args)



"""# NASA Mars Pathfinder Technology Risk DSM

The technology risk DSM (TR-DSM) helps to identify where most of the design effort
should go if one wants to mitigate technological failures and improve robustness of a
system. All components of a system are assigned with a technology risk factor,
which indicates the probability of failure, unprovenness, or uncertainty in design.
The technology risk factor at NASA is loosely defined as the inverse of the technology
readiness level:

| TRF | NASA Technology Readiness Level Definition                                           | TRL |
|-----|--------------------------------------------------------------------------------------|-----|
|  1  | Actual system "flight proven" through successful mission operations                  |  9  |
|  2  | Actual system completed and "flight qualified" through test and demonstration        |  8  |
|  2  | System prototype in a space environment                                              |  7  |
|  3  | System/subsystem model or prototype demonstration in a relevant environment          |  6  |
|  4  | Component and/or breadboard validation in relevant environment                       |  5  |
|  4  | Component and/or breadboard validation in laboratory environment                     |  4  |
|  5  | Analytical and experimental critical function and/or characteristic proof-of-concept |  3  |
|  5  | Technology concept and/or application formulated                                     |  2  |
|  5  | Basic principles observed and reported                                               |  1  |


The dataset includes both an adjacency value as if the DSM were a regular product DSM
with interface strength values and the technology risk value which is computed using:

    TR = TRF-source * TRF-target * adjacency

Reference:
    Brady, Timothy K. 2002. Utilization of Dependency Structure Matrix Analysis to
    Assess Complex Project Designs. Proceedings of ASME Design Engineering Technical
    Conferences, no. DETCZ002/DTM-34031, Montreal, Canada.
"""

node_weights = ["technology risk factor"]
edge_weights = ["adjacency", "technology risk"]



"""# Graph class module

The [`Graph`][ragraph.graph.Graph] class is the core of this package. It is the internal storage
format that facilitates all conversions. These class definitions are designed to
accommodate a wide range of graph formats.
"""
from collections import OrderedDict, defaultdict
from copy import deepcopy
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from uuid import UUID

from ragraph import utils
from ragraph.edge import Edge
from ragraph.generic import Annotations, Metadata
from ragraph.node import Node

try:
    import numpy as np

except ImportError:
    np = None  # type: ignore


class ConsistencyError(Exception):
    pass


class Graph(Metadata):
    """Graph of [`Node`][ragraph.node.Node] objects and [`Edge`][ragraph.edge.Edge] objects between
    them.

    Arguments:
        nodes: Iterable of graph nodes.
        edges: Iterable of graph edges.
        add_parents: Recursively add parent nodes of the provided nodes.
        add_children: Recursively add child nodes of the provided nodes.
        name: Instance name. Given a UUID if none provided.
        kind: Kind or main category of this instance.
        labels: Labels categorizing this instance.
        weights: Dictionary of weights attached to this instance.
        annotations: Miscellaneous properties of this instance.
        uuid: Fixed UUID if desired, generated when left set to None.
    """

    def __init__(
        self,
        nodes: Optional[Iterable[Node]] = None,
        edges: Optional[Iterable[Edge]] = None,
        add_parents: bool = False,
        add_children: bool = False,
        name: Optional[str] = None,
        kind: Optional[str] = None,
        labels: Optional[List[str]] = None,
        weights: Optional[Dict[str, Union[int, float]]] = None,
        annotations: Optional[Union[Annotations, Dict[str, Any]]] = None,
        uuid: Optional[Union[str, UUID]] = None,
    ):
        Metadata.__init__(
            self,
            name=name,
            kind=kind,
            labels=labels,
            weights=weights,
            annotations=annotations,
            uuid=uuid,
        )
        self._nodes: Dict[str, Node] = OrderedDict()
        self._node_uuid_dict: Optional[Dict[UUID, Node]] = None  # Lazy
        self._edges: Dict[str, Edge] = OrderedDict()
        self._edge_uuid_dict: Optional[Dict[UUID, Edge]] = None  # lazy
        self._directed_edges: defaultdict = defaultdict(lambda: defaultdict(list))
        self._reversed_edges: defaultdict = defaultdict(lambda: defaultdict(list))
        self.add_parents = add_parents
        self.add_children = add_children
        setattr(self, "nodes", nodes)
        setattr(self, "edges", edges)

    def __str__(self) -> str:
        nodes, edges = self.node_count, self.edge_count
        ns, es = "" if nodes == 1 else "s", "" if edges == 1 else "s"
        counts = f"{nodes} node{ns}, {edges} edge{es}"
        return f"{self.__class__.__name__}(name={self.name}, {counts})"

    def __repr__(self) -> str:
        kwargdict = dict(
            name=repr(self.name),
            kind=repr(self.kind),
            labels=self.labels,
            weights=self.weights,
            annotations=self.annotations,
            uuid=repr(self.uuid),
        )
        kwargs = ", ".join([str(k) + "=" + str(v) for k, v in kwargdict.items()])
        nodes, edges = self.node_count, self.edge_count
        ns, es = "" if nodes == 1 else "s", "" if edges == 1 else "s"
        counts = f"{nodes} node{ns}, {edges} edge{es}"
        return "<{}.{}({}), {} at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            kwargs,
            counts,
            hex(id(self)),
        )

    def __eq__(self, other) -> bool:
        return self.as_dict(use_uuid=False) == other.as_dict(use_uuid=False)

    def __getitem__(self, key: Union[str, UUID, Tuple[str, str]]):
        if isinstance(key, str):
            return self._nodes[key]

        if isinstance(key, (tuple, list)) and len(key) == 2:
            return self._directed_edges[key[0]][key[1]]

        if isinstance(key, UUID):
            if key in self.node_uuid_dict:
                return self.node_uuid_dict[key]
            if key in self.edge_uuid_dict:
                return self.edge_uuid_dict[key]
            raise KeyError(f"Cannot find any Node or Edge with UUID '{key}'.")

        raise TypeError(
            f"Key '{key}' should be a node name string or "
            'a ("source", "target") tuple of strings.'
        )

    def __contains__(self, key: Union[str, UUID, Tuple[str, str]]):
        try:
            return bool(self.__getitem__(key))
        except KeyError:
            return False

    def __copy__(self):
        cp = Metadata.__copy__(self)
        cp._node_uuid_dict = None
        cp._edge_uuid_dict = None
        return cp

    def __deepcopy__(self, memo):
        cp = Metadata.__deepcopy__(self, memo)
        cp._node_uuid_dict = None
        cp._edge_uuid_dict = None
        return cp

    @property
    def nodes(self) -> List[Node]:
        """Nodes as a list."""
        return self.node_list

    @nodes.setter
    def nodes(self, value: Optional[Iterable[Node]]):
        self._nodes = OrderedDict()
        if isinstance(value, dict):
            value = value.values()
        if value:
            for node in value:
                self.add_node(node)

    @property
    def node_count(self) -> int:
        """Number of nodes in the graph."""
        return len(self._nodes)

    @property
    def node_list(self) -> List[Node]:
        """Nodes as a list."""
        return list(self._nodes.values())

    @property
    def node_dict(self) -> Dict[str, Node]:
        """Node dictionary by node name."""
        return self._nodes

    @property
    def node_uuid_dict(self) -> Dict[UUID, Node]:
        """Node dictionary by node UUID."""
        if self._node_uuid_dict is None:
            self._node_uuid_dict = {n.uuid: n for n in self.node_gen}
        return self._node_uuid_dict

    @property
    def node_gen(self) -> Generator[Node, None, None]:
        """Yield nodes."""
        yield from self._nodes.values()

    @property
    def edges(self) -> List[Edge]:
        """Edges as a list."""
        return self.edge_list

    @edges.setter
    def edges(self, value: Optional[Iterable[Edge]]):
        self._edges = OrderedDict()
        self._directed_edges = defaultdict(lambda: defaultdict(list))
        self._reversed_edges = defaultdict(lambda: defaultdict(list))
        if isinstance(value, dict):
            for source in value:
                for target in value[source]:
                    for edge in value[source][target]:
                        self.add_edge(edge)
        elif value:
            for edge in value:
                self.add_edge(edge)

    @property
    def edge_list(self) -> List[Edge]:
        """Edges as a list."""
        return list(self._edges.values())

    @property
    def edge_dict(self) -> Dict[str, Edge]:
        """Edge dictionary by edge name."""
        return self._edges

    @property
    def edge_uuid_dict(self) -> Dict[UUID, Edge]:
        """Edge dictionary by node UUID."""
        if self._edge_uuid_dict is None:
            self._edge_uuid_dict = {e.uuid: e for e in self.edge_gen}
        return self._edge_uuid_dict

    @property
    def directed_edges(self) -> Dict[str, Dict[str, List[Edge]]]:
        """Nested edge dictionary from source name to target name to a list of edges."""
        return self._directed_edges

    @property
    def reversed_edges(self) -> Dict[str, Dict[str, List[Edge]]]:
        """Nested edge dictionary from target name to source name to a list of edges."""
        return self._reversed_edges

    @property
    def edge_gen(self) -> Generator[Edge, None, None]:
        """Yield edges."""
        for source in self._directed_edges:
            for target in self._directed_edges[source]:
                yield from self._directed_edges[source][target]

    @property
    def edge_count(self) -> int:
        """Number of edges in the graph."""
        return len(self._edges)

    def add_node(self, node: Node):
        """Add a new node to this graph.

        If the node instance already exists in the graph, it is ignored.

        Arguments:
            node: Node to add.
        """
        if not isinstance(node, Node):
            raise TypeError("%r is not an instance of Node but of %r." % (node, type(node)))

        # Append own node if not exists.
        if node.name in self._nodes:
            if node == self._nodes[node.name]:
                return
            raise ValueError(f"Another Node named '{node.name}' already exists in this Graph.")
        if node.uuid in self.node_uuid_dict:
            if node == self.node_uuid_dict[node.uuid]:
                return
            raise ValueError(f"Another Node with UUID '{node.uuid}' already exists in this Graph.")
        self._nodes[node.name] = node
        self.node_uuid_dict[node.uuid] = node

        # Append parent
        if self.add_parents and node.parent:
            self.add_node(node.parent)

        # Append children
        if self.add_children and node.children:
            for child in node.children:
                self.add_node(child)

    def del_node(self, node: Union[Node, str, UUID], inherit: bool = False):
        """Remove a node and related edges from the graph by name or Node instance.

        Arguments:
            node: Node to remove.
            inherit: Whether to inherit hierarchy for underlying nodes to the parent.
        """
        if isinstance(node, str):
            node = self._nodes[node]
        elif isinstance(node, UUID):
            node = self.node_uuid_dict[node]

        if inherit and node.parent and node.children:
            for child in node.children:  # Is bus does not carry over to new parent.
                child.is_bus = False
            node.parent.children = node.parent.children + node.children
        node.children = []
        node.parent = None
        for e in list(self.edges_from(node)):
            self.del_edge(e)
        for e in list(self.edges_to(node)):
            self.del_edge(e)
        self.node_uuid_dict.pop(node.uuid)
        self._nodes.pop(node.name)

    def add_edge(self, edge: Edge):
        """Add a new edge to the graph. Source and target need to exist in the graph.

        Arguments:
            edge: Edge to add.

        Raises:
            ValueError: if the source or target doesn't exist or the edge instance already does.
        """
        source = edge.source.name
        target = edge.target.name
        if source not in self.node_dict:
            raise ValueError("Source node %r does not exist in graph." % edge.source)
        if target not in self.node_dict:
            raise ValueError("Target node %r does not exist in graph." % edge.target)
        if edge.name in self._edges:
            if edge == self._edges[edge.name]:
                return
            raise ValueError(f"Edge name '{edge.name}' already exists in the graph.")
        if edge.uuid in self.edge_uuid_dict:
            if edge == self.edge_uuid_dict[edge.uuid]:
                return
            raise ValueError(f"Edge UUID '{edge.name}' already exists in the graph.")
        self._edges[edge.name] = edge
        self.edge_uuid_dict[edge.uuid] = edge
        self._directed_edges[source][target].append(edge)
        self._reversed_edges[target][source].append(edge)

    def del_edge(self, edge: Union[Edge, str, UUID]):
        """Remove an edge from the graph."""
        if isinstance(edge, str):
            edge = self._edges[edge]
        elif isinstance(edge, UUID):
            edge = self.edge_uuid_dict[edge]

        self.edge_uuid_dict.pop(edge.uuid)
        self._edges.pop(edge.name)

        source = edge.source.name
        target = edge.target.name

        self._directed_edges[source][target].remove(edge)
        if not self._directed_edges[source][target]:
            self._directed_edges[source].pop(target)

        self._reversed_edges[target][source].remove(edge)
        if not self._reversed_edges[source][target]:
            self._reversed_edges[source].pop(target)

    @property
    def node_kinds(self) -> List[str]:
        """List of unique node kinds in the graph."""
        return self.get_kinds(self.node_gen)

    @property
    def edge_kinds(self) -> List[str]:
        """List of unique edge kinds in the graph."""
        return self.get_kinds(self.edge_gen)

    def get_kinds(self, elements: Union[Iterable[Node], Iterable[Edge]]) -> List[str]:
        """Get an alphabetically sorted list of unique node/edge kinds in the graph."""
        return sorted(set(elem.kind for elem in elements), key=str)

    def get_nodes_by_kind(self, kind: str) -> List[Node]:
        """Get nodes/edges by kind."""
        return [node for node in self.node_gen if node.kind == kind]

    def get_edges_by_kind(self, kind: str) -> List[Edge]:
        """Get edges by kind."""
        return [edge for edge in self.edge_gen if edge.kind == kind]

    @property
    def node_weight_labels(self) -> List[str]:
        """List of unique node weight labels in the graph."""
        return self.get_weight_labels(self.node_gen)

    @property
    def edge_weight_labels(self) -> List[str]:
        """List of unique edge weight labels in the graph."""
        return self.get_weight_labels(self.edge_gen)

    def get_weight_labels(self, elements: Union[Iterable[Node], Iterable[Edge]]) -> List[str]:
        """Get an alphabetically sorted list of unique labels of node/edge weights."""
        labels: Set[str] = set()
        for elem in elements:
            labels.update(elem.weights.keys())
        return sorted(labels, key=str)

    @property
    def node_labels(self) -> List[str]:
        """Alphabetically sorted list of unique node labels in the graph."""
        labels = {label for node in self.node_gen for label in node.labels}
        return sorted(labels, key=str)

    @property
    def edge_labels(self) -> List[str]:
        """Alphabetically sorted list of unique edge labels in the graph."""
        labels = {label for edge in self.edge_gen for label in edge.labels}
        return sorted(labels, key=str)

    @property
    def roots(self) -> List[Node]:
        """List of roots in the graph."""
        return [node for node in self.node_gen if node.is_root]

    @property
    def leafs(self) -> List[Node]:
        """List of roots in the graph."""
        return [node for node in self.node_gen if node.is_leaf]

    @property
    def max_depth(self) -> int:
        """Maximum node depth in the graph."""
        return max([node.depth for node in self.leafs], default=0)

    def edges_from(self, source: Union[str, Node]) -> Generator[Edge, None, None]:
        """Yield edges originating from a given source node."""
        if isinstance(source, Node):
            source = source.name
        for target in self._directed_edges[source]:
            yield from self._directed_edges[source][target]

    def targets_of(self, source: Union[str, Node]) -> Generator[Node, None, None]:
        """Yield target nodes of a given source node."""
        if isinstance(source, Node):
            source = source.name
        yield from [
            self._nodes[target]
            for target in self._directed_edges[source]
            if self._directed_edges[source][target]
        ]

    def edges_to(self, target: Union[str, Node]) -> Generator[Edge, None, None]:
        """Yield edges towards a given target node."""
        if isinstance(target, Node):
            target = target.name
        for source in self._reversed_edges[target]:
            yield from self._reversed_edges[target][source]

    def sources_of(self, target: Union[str, Node]) -> Generator[Node, None, None]:
        """Yield source nodes of a given target node."""
        if isinstance(target, Node):
            target = target.name
        yield from [
            self._nodes[source]
            for source in self._reversed_edges[target]
            if self._reversed_edges[target][source]
        ]

    def edges_between(
        self,
        source: Union[str, Node],
        target: Union[str, Node],
        inherit: bool = False,
        loops: bool = True,
    ) -> Generator[Edge, None, None]:
        """Yield edges between source and target node.

        Arguments:
            source: Source node.
            target: Target node.
            inherit: Whether to include edges between descendants of given nodes.
            loops: Whether to include self-loop edges.
        """
        if source == target and not loops:
            return
        source_node = source if isinstance(source, Node) else self._nodes[source]
        target_node = target if isinstance(target, Node) else self._nodes[target]
        sources, targets = [source_node.name], [target_node.name]
        if inherit:
            sources.extend(n.name for n in source_node.descendant_gen)
            targets.extend(n.name for n in target_node.descendant_gen)

        for source in sources:
            for target in targets:
                if source == target and not loops:
                    continue
                yield from self._directed_edges[source][target]

    def edges_between_all(
        self,
        sources: Iterable[Union[str, Node]],
        targets: Iterable[Union[str, Node]],
        inherit: bool = False,
        loops: bool = True,
    ) -> Generator[Edge, None, None]:
        """Yield all edges between a set of source and a set of target nodes.

        Arguments:
            sources: Source nodes.
            targets: Target nodes.
            inherit: Whether to include edges between descendants of given nodes.
            loops: Whether to include self-loop edges.
        """
        for source in sources:
            for target in targets:
                yield from self.edges_between(source, target, inherit=inherit, loops=loops)

    def get_graph_slice(
        self,
        nodes: Optional[Iterable[Node]] = None,
        edges: Optional[Iterable[Edge]] = None,
        inherit: bool = False,
    ) -> "Graph":
        """Return the graph slice with only the given nodes and the edges between those.

        Arguments:
            nodes: Optional set of nodes to include in the new deep copy.
                Defaults to all.
            edges: Optional set of edges to include in the new deep copy.
                Defaults to all between selected nodes.
            inherit: Whether to include child nodes of the given nodes and edges
                between them (unless edges are explicitly specified).

        Note:
            This returns a deepcopy, so the copied nodes and edges do NOT share any
            changes! All parent and child nodes not in the list are excluded.
        """
        nodes = self.nodes if nodes is None else list(nodes)
        edges = self.edges_between_all(nodes, nodes, inherit) if edges is None else list(edges)

        # Get a deepcopy to prevent changing nodes in the current graph instance.
        graph = deepcopy(Graph(nodes=nodes, edges=edges, add_children=inherit))

        # Reinstate relationships only between given nodes.
        graph_nodes = graph.nodes
        for node in graph_nodes:
            if node.parent:
                node.parent = graph[node.parent.name] if node.parent in graph.nodes else None
            node.children = [c for c in node.children if c in graph.nodes]
            if node.is_bus and not node.parent:
                node.is_bus = False

        return graph

    def get_node_selection(
        self,
        node_kinds: Optional[List[str]] = None,
        edge_kinds: Optional[List[str]] = None,
        depth: int = 2,
        selection_mode: str = "dependent",
    ) -> List[Node]:
        """Select specific nodes from this graph in a structured order.

        Arguments:
            node_kinds: The kind of nodes to be selected.
            edge_kinds: The kind of edges to be selected.
            depth: The maximum depth of node to be selected.
            selection_mode: The selection mode. Either 'dependent' or 'independent'.

        Note:
            The selection mode argument determines how nodes of different kinds are
            selected. If the selection mode is set to 'dependent', the first node kind
            in the `node_kinds` list is considered to be the 'lead node kind'.
            Nodes of different kind than the lead node kind, are only selected if they
            have a dependency with at least one of the selected nodes of the lead node
            kind. If the selection mode is set to 'independent' this dependency
            condition is dropped.
        """
        nkinds = node_kinds or self.node_kinds
        ekinds = edge_kinds or self.edge_kinds
        return utils.select_nodes(self, nkinds, ekinds, depth, selection_mode)

    def get_edge_selection(
        self, nodes: Optional[List[Node]] = None, edge_kinds: Optional[List[str]] = None
    ) -> List[Edge]:
        """Select specific edges from this graph.

        Arguments:
            nodes: The list of nodes between which the edges must be selected.
            edge_kinds: The kinds of edges to be selected.

        Returns:
            List of selected edges.
        """

        nlist = nodes or self.get_node_selection(edge_kinds=edge_kinds)
        ekinds = edge_kinds or self.edge_kinds
        ekinds = sorted(set(ekinds))

        return [e for e in self.edges_between_all(sources=nlist, targets=nlist) if e.kind in ekinds]

    def get_node_and_edge_selection(
        self,
        node_kinds: Optional[List[str]] = None,
        edge_kinds: Optional[List[str]] = None,
        depth: int = 2,
        selection_mode: str = "dependent",
    ) -> Tuple[List[Node], List[Edge]]:
        """Select specific nodes and edges from this graph in a structured order.

        Arguments:
            node_kinds: The kind of nodes to be selected.
            edge_kinds: The kind of edges to be selected.
            depth: The maximum depth of node to be selected.
            selection_mode: The selection mode. Either 'dependent' or 'independent'.

        Note:
            The selection mode argument determines how nodes of different kinds are
            selected. If the selection mode is set to 'dependent', the first node kind
            in the `node_kinds` list is considered to be the 'lead node kind'.
            Nodes of different kind than the lead node kind, are only selected if they
            have a dependency with at least one of the selected nodes of the lead node
            kind. If the selection mode is set to 'independent' this dependency
            condition is dropped.
        """

        nodes = self.get_node_selection(
            node_kinds=node_kinds,
            edge_kinds=edge_kinds,
            depth=depth,
            selection_mode=selection_mode,
        )

        edges = self.get_edge_selection(nodes=nodes, edge_kinds=edge_kinds)

        return nodes, edges

    def get_hierarchy_dict(self, roots: Optional[List[Node]] = None, levels: Optional[int] = None):
        """Return a dictionary of the (partial) hierarchical node structure.

        Arguments:
            roots: Root nodes of the hierarchy to calculate.
            levels: Number of levels to include below the roots.
        """
        if roots is None:
            roots = self.roots

        if levels is None:
            return {root.name: self.get_hierarchy_dict(root.children) for root in roots}
        elif levels > 0:
            return {
                root.name: self.get_hierarchy_dict(root.children, levels=levels - 1)
                for root in roots
            }

        return {root.name: dict() for root in roots}

    def get_adjacency_matrix(
        self,
        nodes: Optional[Union[List[Node], List[str]]] = None,
        inherit: bool = False,
        loops: bool = False,
        only: Optional[List[str]] = None,
    ) -> Union["np.ndarray", List[List[float]]]:
        """Convert graph data into a numerical adjacency matrix.

        Arguments:
            nodes: Optional list of nodes for which to get the adjacency matrix.
            inherit: Whether to count weights between children of the given nodes.
            loops: Whether to calculate self-loops from a node to itself.
            only: Optional subset of edge weights to consider. See [`Edge`][ragraph.edge.Edge] for
                default edge weight implementation.

        Returns:
            Adjacency matrix as a 2D numpy array if numpy is present. Otherwise it will
            return a 2D nested list.
        """
        from ragraph.io.matrix import to_matrix

        return to_matrix(self, rows=nodes, cols=nodes, inherit=inherit, loops=loops, only=only)

    def get_mapping_matrix(
        self,
        rows: Optional[Union[List[Node], List[str]]] = None,
        cols: Optional[Union[List[Node], List[str]]] = None,
        inherit: bool = False,
        loops: bool = False,
        only: Optional[List[str]] = None,
    ) -> Union["np.ndarray", List[List[float]]]:
        """Convert graph data into a numerical mapping matrix.

        Arguments:
            rows: Nodes representing the matrix rows.
            cols: Nodes representing the matrix columns if different from the rows.
            inherit: Whether to count weights between children of the given nodes.
            loops: Whether to calculate self-loops from a node to itself.
            only: Optional subset of edge weights to consider. See [`Edge`][ragraph.edge.Edge] for
                default edge weight implementation.

        Returns:
            Adjacency matrix as a 2D numpy array if numpy is present. Otherwise it will
            return a 2D nested list.
        """
        from ragraph.io.matrix import to_matrix

        return to_matrix(self, rows=rows, cols=cols, inherit=inherit, loops=loops, only=only)

    def get_ascii_art(
        self,
        nodes: Optional[List[Node]] = None,
        edge: str = "X",
        diag: str = "\u25A0",
        show: bool = False,
    ) -> Optional[str]:
        """Get a unicode ASCII art representation of a binary adjacency matrix for a
        given set of nodes.

        Arguments:
            nodes: Nodes to include in the art.
            edge: Mark to use when there's an edge between nodes. (X by default)
            diag: Mark to use on the diagonal.
            show: Whether to directly print the ASCII art.

        Returns:
            ASCII art representation of this graph.
        """
        if nodes is None:
            nodes = self.leafs
        if not nodes:
            raise ValueError("Empty graph, cannot create ASCII art.")

        namelens = [len(n.name) for n in nodes]
        maxlen = max(namelens)
        dim = len(nodes)

        ddd = "\u2500\u2500\u2500"  # ---
        pad = maxlen * " "

        # Grid lines
        topline = pad + "\u250C" + (dim - 1) * (ddd + "\u252C") + ddd + "\u2510"
        sepline = pad + "\u251C" + (dim - 1) * (ddd + "\u253C") + ddd + "\u2524"
        endline = pad + "\u2514" + (dim - 1) * (ddd + "\u2534") + ddd + "\u2518"

        lines = [topline]

        for i, tgt in enumerate(nodes):
            line = (maxlen - namelens[i]) * " " + tgt.name + "\u2525"
            for j, src in enumerate(nodes):
                if i == j:
                    mark = diag
                elif self._directed_edges.get(src.name, dict()).get(tgt.name, False):
                    mark = edge
                else:
                    mark = " "
                line += " {} \u2502".format(mark)

            lines.append(line)
            lines.append(sepline)
        lines[-1] = endline
        txt = "\n".join(lines)
        if show:
            print(txt)
            return None
        else:
            return txt

    @property
    def json_dict(self) -> Dict[str, Any]:
        """JSON dictionary representation."""
        return self.as_dict(use_uuid=True)

    def as_dict(self, use_uuid: bool = False) -> Dict[str, Any]:
        """Return a copy as a (serializable) dictionary.

        Arguments:
            use_uuid: Whether to use UUIDs instead of names.

        Returns:
            nodes: Node names or UUIDs to Node dictionaries.
            kind: Kind as str.
            labels: Labels as list of str.
            weights: Weights as dict.
            annotations: Annotations as a dictionary.
            uuid: UUID as str if toggled.

        """
        if use_uuid:
            return dict(
                nodes={str(n.uuid): n.json_dict for n in self.nodes},
                edges={str(e.uuid): e.json_dict for e in self.edges},
                name=self.name,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
                uuid=str(self.uuid),
            )
        else:
            return dict(
                nodes={n.name: n.as_dict(use_uuid=False) for n in self.nodes},
                edges={e.name: e.as_dict(use_uuid=False) for e in self.edges},
                name=self.name,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
            )

    def check_consistency(self, raise_error: bool = True) -> bool:
        """Check the consistency of this graph.

        Arguments:
            raise_error: Whether to raise an error instead of returning a bool.

        Returns:
            Whether nodes aren't their own ancestor/descendant and if all their children
            and parents exist in the graph. Raises an error for inconsistencies if
            `raise_error` is set to `True`.
        """
        consistent = True

        try:
            for node in self.node_list:
                a = node
                while a.parent:
                    if a.parent == node:
                        raise ConsistencyError("Node {} is in its own ancestors.".format(node))
                    a = a.parent
                ds = node.children
                while ds:
                    if node in ds:
                        raise ConsistencyError("Node {} is in its own descendants.".format(node))
                    ds = [c for d in ds for c in d.children]
                if node.parent and node.parent not in self.node_list:
                    raise ConsistencyError(
                        "Node {}'s parent {} is missing in the graph.".format(node, node.parent)
                    )
                if node.parent and node not in node.parent.children:
                    raise ConsistencyError(
                        "Node {}'s does not occur in parent {}'s children.".format(
                            node, node.parent
                        )
                    )
                for child in node.children:
                    if child not in self.node_list:
                        raise ConsistencyError(
                            "Node {}'s child {} is missing in the graph.".format(node, child)
                        )
                    if child.parent != node:
                        raise ConsistencyError(
                            "Node {}'s child has a different parent {}.".format(node, child.parent)
                        )
                if node.is_bus and not node.parent:
                    raise ConsistencyError(
                        "Node {} is a bus node, but has no parent to be it for.".format(node)
                    )

            for edge in self.edge_list:
                if edge.source.name not in self.node_dict:
                    raise ConsistencyError(f"Edge source not in Graph: {repr(edge.source)}")
                if edge.target.name not in self.node_dict:
                    raise ConsistencyError(f"Edge target not in Graph: {repr(edge.source)}")

        except ConsistencyError as e:
            consistent = False
            if raise_error:
                raise e

        return consistent


class GraphView(Graph):
    """A view on a [`Graph`][ragraph.graph.Graph] object. It works exactly like the
    [`Graph`][ragraph.graph.Graph] object, except for that the nodes and edges are filtered
    according to the view function.

    Arguments:
        graph: Graph to provide a view on.
        view_func: View function that filters the [`Graph`][ragraph.graph.Graph]'s nodes and edges.
        view_kwargs: Optional arguments to pass to the view function.
    """

    def __init__(
        self,
        graph: Graph,
        view_func: Optional[
            Callable[
                [Any],
                Tuple[
                    Union[Iterable[Node], Dict[str, Node]],
                    Union[Iterable[Edge], Dict[str, Edge]],
                ],
            ]
        ] = None,
        view_kwargs: Optional[Dict[str, Any]] = None,
    ):
        Metadata.__init__(
            self,
            name=graph.name,
            kind=graph.kind,
            labels=graph.labels,
            weights=graph.weights,
            annotations=graph.annotations,
        )
        self._graph = graph
        self._view_func = view_func
        self._view_kwargs = view_kwargs

        self._vnodes: Optional[Dict[str, Node]] = None
        self._vnode_uuid_dict: Optional[Dict[UUID, Node]] = None
        self._vedges: Optional[Dict[str, Edge]] = None
        self._vedge_uuid_dict: Optional[Dict[UUID, Edge]] = None
        self._vdirected_edges: Optional[defaultdict] = None
        self._vreversed_edges: Optional[defaultdict] = None
        self.update()

    @property
    def graph(self) -> Graph:
        """Graph to provide a view on."""
        return self._graph

    @graph.setter
    def graph(self, value: Graph):
        self._graph = value
        self.reset()

    @property
    def view_func(
        self,
    ) -> Optional[
        Callable[
            [Any],
            Tuple[
                Union[Iterable[Node], Dict[str, Node]],
                Union[Iterable[Edge], Dict[str, Edge]],
            ],
        ]
    ]:
        """View function that filters the [`Graph`][ragraph.graph.Graph]'s nodes and edges."""
        return self._view_func

    @view_func.setter
    def view_func(
        self,
        value: Optional[
            Callable[
                [Any],
                Tuple[
                    Union[Iterable[Node], Dict[str, Node]],
                    Union[Iterable[Edge], Dict[str, Edge]],
                ],
            ]
        ],
    ):
        self._view_func = value
        self.reset()

    @property
    def view_kwargs(self) -> Dict[str, Any]:
        """Optional arguments to pass to the view function."""
        if self._view_kwargs is None:
            return dict()
        return self._view_kwargs

    @view_kwargs.setter
    def view_kwargs(self, value: Dict[str, Any]):
        self._view_kwargs = value
        self.reset()

    def update(self, **kwargs):
        """Apply the set view function with arguments."""
        # Just proxy if no view function is set.
        if self.view_func is None:
            return self.proxy()

        # Get view filter results.
        view_kwargs = kwargs or self.view_kwargs
        nodes, edges = self.view_func(self.graph, **view_kwargs)

        # Cast to dicts.
        ndict = nodes if isinstance(nodes, dict) else {n.name: n for n in nodes}
        edict = edges if isinstance(edges, dict) else {e.name: e for e in edges}

        # Populate directed/reversed edges.
        directed_edges = defaultdict(lambda: defaultdict(list))
        reversed_edges = defaultdict(lambda: defaultdict(list))
        for e in edict.values():
            directed_edges[e.source.name][e.target.name].append(e)
            reversed_edges[e.target.name][e.source.name].append(e)

        # Update private views.
        self._vnodes, self._vedges = ndict, edict
        self._vnode_uuid_dict = {n.uuid: n for n in ndict.values()}
        self._vedge_uuid_dict = {e.uuid: e for e in edict.values()}
        self._vdirected_edges, self._vreversed_edges = directed_edges, reversed_edges

        # Save view kwargs after view succeeded.
        self._view_kwargs = view_kwargs

    def proxy(self):
        """Undo any filtering and just set this view to proxy the Graph 1:1."""
        self._vnodes, self._vedges = self.graph._nodes, self.graph._edges
        self._vnode_uuid_dict = self.graph.node_uuid_dict
        self._vedge_uuid_dict = self.graph.edge_uuid_dict
        self._vdirected_edges = self.graph._directed_edges
        self._vreversed_edges = self.graph._reversed_edges

    def reset(self):
        """Reset the private properties after the original graph might have changed."""
        self._vnodes = None
        self._vnode_uuid_dict = None
        self._vedges = None
        self._vedge_uuid_dict = None
        self._vdirected_edges = None
        self._vreversed_edges = None

    @property  # type: ignore
    def _nodes(self) -> Optional[Dict[str, Node]]:  # type: ignore
        if self._vnodes is None:
            self.update()
        return self._vnodes

    @_nodes.setter
    def _nodes(self, value: Dict[str, Node]):
        self.graph._nodes = value
        self.reset()

    @property  # type: ignore
    def node_uuid_dict(self) -> Optional[Dict[UUID, Node]]:  # type: ignore
        if self._vnode_uuid_dict is None:
            self.update()
        return self._vnode_uuid_dict

    @property  # type: ignore
    def _edges(self) -> Optional[Dict[str, Edge]]:  # type: ignore
        if self._vedges is None:
            self.update()
        return self._vedges

    @_edges.setter
    def _edges(self, value: Dict[str, Edge]):
        self.graph._edges = value
        self.reset()

    @property  # type: ignore
    def edge_uuid_dict(self) -> Optional[Dict[UUID, Edge]]:  # type: ignore
        if self._vedge_uuid_dict is None:
            self.update()
        return self._vedge_uuid_dict

    @property  # type: ignore
    def _directed_edges(self) -> defaultdict:  # type: ignore
        if self._vdirected_edges is None:
            self.update()
        assert self._vdirected_edges is not None
        return self._vdirected_edges

    @_directed_edges.setter
    def _directed_edges(self, value: defaultdict):
        self.graph._directed_edges = value
        self.reset()

    @property  # type: ignore
    def _reversed_edges(self) -> defaultdict:  # type: ignore
        if self._vreversed_edges is None:
            self.update()
        assert self._vreversed_edges is not None
        return self._vreversed_edges

    @_reversed_edges.setter
    def _reversed_edges(self, value: defaultdict):
        self.graph._reversed_edges = value
        self.reset()

    def add_node(self, node: Node):
        self.graph.add_node(node)
        self.reset()

    def del_node(self, node: Union[Node, str, UUID], inherit: bool = False):
        self.graph.del_node(node, inherit=inherit)
        self.reset()

    def add_edge(self, edge: Edge):
        self.graph.add_edge(edge)
        self.reset()

    def del_edge(self, edge: Union[Edge, str, UUID]):
        self.graph.del_edge(edge)
        self.reset()



"""# Integral architecture example

Graph describing a design problem using objects that need to be designed based on
certain aspects. It is part of a set of three design problem graphs, each ideally solved
with a different approach -- being a modular, **integral** or mixed architecture.

This particular graph describes a design problem which is ideally solved using a
**integral object architecture**. That the architecture consists vertically integrated
modules. That is, each module in one domain integrates with only one (or very few) in
the next and there is no crossover (mixing and matching across domains like with a
modular approach). This means that interfaces between modules in different domains are
relatively few. A module's design solution should be applicable to every member.

Nodes are of kind "object" or "aspect_1", "aspect_2", and "aspect_3". The numbers
represent different aspect domains. An "incidence" kind edge from an object to an aspect
means that it possesses that aspect.

Authors:
    Tim Wilschut and Tiemen Schuijbroek, October 2019.
"""



"""# Delta analysis"""


from typing import Type

from ragraph.analysis.comparison.utils import (
    EdgeDescriptor,
    EdgeDescriptorLike,
    NodeDescriptor,
    NodeDescriptorLike,
    TagMode,
    tag,
)
from ragraph.graph import Graph


def delta_graph(
    graph_a: Graph,
    graph_b: Graph,
    delta_a: str = "delta_a",
    delta_b: str = "delta_b",
    common: str = "common",
    tag_mode: TagMode = TagMode.KIND,
    node_descriptor: Type[NodeDescriptorLike] = NodeDescriptor,
    edge_descriptor: Type[EdgeDescriptorLike] = EdgeDescriptor,
) -> Graph:
    """Get the delta graph between two graphs.

    Arguments:
        graph_a: First graph to compare.
        graph_b: Second graph to compare.
        delta_a: Name for nodes and edges unique to the first graph.
        delta_b: Name for nodes and edges unique to the second graph.
        common: Name for the common nodes and edges occurring in both graphs.

    Note:
        Graphs are compared on leaf node level and compared by name only.
        Please provide appropriate input as such and look to
        [`get_graph_slice`][ragraph.graph.Graph.get_graph_slice] for example on how to obtain a
        subgraph to your liking.
    """
    graph = Graph(name="delta", kind="delta")

    # Compare leaf nodes by name. Set to unique A first and change to common if found again.
    nds_a = set(node_descriptor.from_node(n) for n in graph_a.leafs)
    nds_b = set(node_descriptor.from_node(n) for n in graph_b.leafs)
    for n in nds_a.difference(nds_b):
        node = n.to_node(graph)
        tag(node, delta_a, tag_mode)
    for n in nds_b.difference(nds_a):
        node = n.to_node(graph)
        tag(node, delta_b, tag_mode)
    for n in nds_a.intersection(nds_b):
        node = n.to_node(graph)
        tag(node, common, tag_mode)

    # Compare edges between the nodes in the newly created graph.
    eds_a = set(
        edge_descriptor.from_edge(e)
        for e in graph_a.edges_between_all(graph_a.leafs, graph_a.leafs, inherit=False, loops=True)
    )
    eds_b = set(
        edge_descriptor.from_edge(e)
        for e in graph_b.edges_between_all(graph_b.leafs, graph_b.leafs, inherit=False, loops=True)
    )
    for e in eds_a.difference(eds_b):
        edge = e.to_edge(graph)
        tag(edge, delta_a, tag_mode)
    for e in eds_b.difference(eds_a):
        edge = e.to_edge(graph)
        tag(edge, delta_b, tag_mode)
    for e in eds_a.intersection(eds_b):
        edge = e.to_edge(graph)
        tag(edge, common, tag_mode)

    return graph



"""# YAML format support"""
from pathlib import Path
from typing import Optional, Union

from ragraph.graph import Graph
from ragraph.io.json import graph_from_json_dict

try:
    from ruamel_yaml import YAML
except ImportError:
    from ruamel.yaml import YAML

yaml = YAML(typ="safe")


def from_yaml(path: Optional[Union[str, Path]] = None, enc: Optional[str] = None) -> Graph:
    """Decode YAML file or string into a Graph.

    Arguments:
        path: YAML file path.
        enc: YAML encoded string.

    Returns:
        Graph object.
    """
    if path is None and enc is None:
        raise ValueError("`path` and `enc` arguments cannot both be `None`.")
    if path is not None and enc is not None:
        raise ValueError("`path` and `enc` arguments cannot both be set.")

    if path:
        json_dict = yaml.load(Path(path))
    else:
        json_dict = yaml.load(enc)

    return graph_from_json_dict(json_dict)


def to_yaml(graph: Graph, path: Optional[Union[str, Path]] = None) -> Optional[str]:
    """Encode Graph to YAML file or string.

    Arguments:
        path: Optional file path to write YAML to.

    Returns:
        YAML string.
    """
    stream = Path(path) if path else None
    return yaml.dump(graph.json_dict, stream=stream)



"""# Binary Decision Diagram compatibility calculation

Binary Decision Diagram (BDD) implementation of the compatibility problem.
"""
from typing import Generator, List, Tuple, Union

import dd.autoref as _bdd
import numpy as np


def _yield_ranges(nums: List[int], start: int = 0) -> Generator[range, None, None]:
    """Generate matrix index ranges."""
    for i in nums:
        yield range(start, start + i)
        start += i


def _recursive(operation: str, bdd: _bdd.BDD, items: List[_bdd.Function]) -> _bdd.Function:
    """Construct a recursive operation function over all items."""
    if len(items) == 1:
        return items[0]
    return bdd.apply(operation, items[0], _recursive(operation, bdd, items[1:]))


def yield_feasible_configurations(
    compat: Union["np.ndarray", List[List[Union[bool, int]]]],
    comp_variant_nums: List[int],
) -> Generator[Tuple[int, ...], None, None]:
    """Get the feasible configurations based on a compatibility matrix between different
    variants of elements and the "bucket" sizes (number of variants of each element).

    Arguments:
        compat: Compatibility matrix (square) between different variants of elements.
            Size is determined by the total number of variants.
        comp_variant_nums: The number of variants for each element. The matrix has to
            be sorted accordingly.

    Returns:
        Feasible configurations as a tuple with a variant's (absolute) index for
        each element.
    """
    bdd = _bdd.BDD()
    dim = len(compat)

    # All constraints that should be met.
    constraints = []

    # Variant toggles
    toggles = [bdd.add_var(i) for i in range(dim)]

    # Variant incompatibilities
    for i in range(dim):
        row = bdd.var(i)
        for j in range(i + 1, dim):
            if compat[i][j]:
                continue
            col = bdd.var(j)
            constraints.append(bdd.apply("=>", row, ~col))

    # Element picking constraints.
    ranges = _yield_ranges(comp_variant_nums)
    for rng in ranges:
        if rng.stop - rng.start == 1:
            constraints.append(bdd.var(rng.start))
            continue
        # Adding this xor guarantees an uneven number is selected.
        constraints.append(_recursive("xor", bdd, [bdd.var(i) for i in rng]))
        # Adding the ands guarantees no duo is selected.
        ands = [
            bdd.apply("=>", bdd.var(i), ~bdd.var(j))
            for i in range(rng.start, rng.stop - 1)
            for j in range(i + 1, rng.stop)
        ]
        constraints.append(_recursive("and", bdd, ands))

    # Construct system and yield solutions.
    system = _recursive("and", bdd, constraints)
    for config in bdd.pick_iter(system, care_vars=toggles):
        yield tuple(k for k, v in config.items() if v)



"""# Color utilities"""
from typing import Tuple


def rgb1_to_rgb(r: float, g: float, b: float) -> Tuple[int, int, int]:
    """Convert [0.0-1.0] scaled RGB to [0-255]."""
    return round(r * 255), round(g * 255), round(b * 255)


def rgb_to_rgb1(r: int, g: int, b: int) -> Tuple[float, float, float]:
    """Convert [0-255] scaled RGB to [0.0-1.0]."""
    return r / 255, g / 255, b / 255


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """Convert RGB color to hex code."""
    return "#{0:02x}{1:02x}{2:02x}".format(r, g, b)


def hex_to_rgb(h: str) -> Tuple[int, int, int]:
    """Convert hex code to RGB tuple."""
    h = h.lstrip("#")
    n = len(h)
    rgb = tuple(int(h[i : i + n // 3], 16) for i in range(0, n, n // 3))
    return rgb[0], rgb[1], rgb[2]



"""# Elevator system decomposed into 175 components

The Complex Elevator System is described using an undirected graph with multiple dependency types.
It describes a machine-room-less elevator called the 'Kone MonoSpace'. It is designed for low- to
midrise buildings and uses permanent-magnet electric motors. The five defined edge types are
spatial, material, mechanical energy, electrical energy and information. It was published in a
variation of 175 elements and a less granular variation of 45 elements. The less granular variation
collapses a set of pre-defined modules and therefore contains less detail of the system.

Reference:
    Niutanen, V., Hölttä-otto, K., Rahardjo, A., & Stowe, H. M. (2017). Complex Elevator System DSM
    - Case for a DSM Design Sprint. In 19th International dependency and structure modeling
    conference, DSM 2017.
"""

edge_weights = [
    "spatial",
    "material",
    "information",
    "mechanical energy",
    "electrical energy",
    "sum",
    "binary",
]



"""# Mixed architecture example

Graph describing a design problem using objects that need to be designed based on certain aspects.
It is part of a set of three design problem graphs, each ideally solved with a different approach --
being a modular, integral or **mixed architecture**.

This particular graph describes a design problem which is ideally solved using a **mixed object
architecture**. That means that the objects are best put into modules with some, but little
interfaces across domains with limited crossover (mixing and matching like with a modular approach).
Each module should be designed based on all aspects that the objects members possess combined, while
taking into account the shared dependencies with other domains. A module's design solution should be
applicable to every member.

Nodes are of kind "object" or "aspect_1", "aspect_2", and "aspect_3". The numbers represent
different aspect domains. An "incidence" kind edge from an object to an aspect means that it
posesses that aspect.

Authors:
    Tim Wilschut and Tiemen Schuijbroek, October 2019.
"""



"""
# Markov clustering module

Reference:
    Wilschut, T., Etman, L. F. P., Rooda, J. E., & Adan, I. J. B. F. (2017). Multilevel
    Flow-Based Markov Clustering for Design Structure Matrices. Journal of Mechanical
    Design, 139(12), 121402. [DOI: 10.1115/1.4037626](https://doi.org/10.1115/1.4037626)
"""

from typing import List, Optional, Tuple, Union

import numpy as np
from numpy.linalg import LinAlgError

from ragraph.analysis._classes import Bound, ClusterAnalysis, Parameter
from ragraph.analysis._utils import create_parent
from ragraph.graph import Graph, Node

markov_params = {
    p.name: p
    for p in [
        Parameter(
            "alpha",
            int,
            "Expansion coefficient. Adjacency matrix is raised to the power of alpha "
            "in each calculation cycle.",
            default=2,
            lower=Bound(1.0, inclusive=True, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "beta",
            float,
            "Inflation coefficient. Entries within the adjacency matrix are raised to "
            "the power of beta each calculation cycle.",
            default=2.0,
            lower=Bound(1.0, inclusive=False, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "mu",
            float,
            "Decay coefficient (usually 1.5 - 3.5). Influence or dependency of nodes "
            "decays according to this coefficient at each node it passes.",
            default=2.0,
            lower=Bound(1.0, inclusive=False, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "max_iter",
            int,
            "Maximum number of matrix calculation cycles after which convergence is "
            "assumed instead of calculated based on machine precision.",
            default=1000,
            lower=Bound(100, inclusive=True, report="warn"),
            upper=Bound(10000, inclusive=True, report="warn"),
        ),
        Parameter(
            "symmetrize",
            bool,
            "Whether to symmetrize asymmetrical graphs.",
            default=True,
        ),
    ]
}


hierarchical_markov_analysis = ClusterAnalysis(
    "Hierarchical Markov clustering",
    description="Hierarchically cluster the graph in place using Hierarchical Markov "
    + "clustering.",
    parameters=markov_params,
)


@hierarchical_markov_analysis
def hierarchical_markov(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    alpha: int = markov_params["alpha"].default,  # type: ignore
    beta: float = markov_params["beta"].default,  # type: ignore
    mu: float = markov_params["mu"].default,  # type: ignore
    max_iter: int = markov_params["max_iter"].default,  # type: ignore
    symmetrize: bool = markov_params["symmetrize"].default,  # type: ignore
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, Union[List[Node], List[str]]]:
    """docstring stub"""
    cluster_roots: List[Node] = [] if leafs is None else leafs  # type: ignore
    children: List[Node] = []

    while set(cluster_roots) != set(children) and len(cluster_roots) > 1:
        # Previous cluster_roots become children for new cluster level.
        children = cluster_roots
        graph, cluster_roots = markov(
            graph,
            root=None,
            leafs=children,
            inherit=inherit,
            loops=loops,
            edge_weights=edge_weights,
            alpha=alpha,
            beta=beta,
            mu=mu,
            inplace=True,  # Recursive calls may work in the given graph.
            max_iter=max_iter,
        )

    return graph, cluster_roots


markov_analysis = ClusterAnalysis(
    "Markov clustering",
    description="Cluster the graph in place using Markov Clustering.",
    parameters=markov_params,
)


@markov_analysis
def markov(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    alpha: int = markov_params["alpha"].default,  # type: ignore
    beta: float = markov_params["beta"].default,  # type: ignore
    mu: float = markov_params["mu"].default,  # type: ignore
    max_iter: int = markov_params["max_iter"].default,  # type: ignore
    symmetrize: bool = markov_params["symmetrize"],  # type: ignore
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, Union[List[Node], List[str]]]:
    """docstring stub"""
    assert leafs is not None
    mat = graph.get_adjacency_matrix(leafs, inherit=inherit, loops=loops, only=edge_weights)
    assert isinstance(mat, np.ndarray)
    if symmetrize:
        mat = mat + mat.T  # Add transpose to get a guaranteed symmetrical matrix.
    tpm = calculate_tpm(mat, mu)
    if alpha > 1:  # Otherwise algorithm does nothing, then column max -> cluster ID.
        i = 0

        # TPM pruning threshold and equality tolerances.
        rtol = max(max_iter**-2, 1e-15)
        atol = max(max_iter**-3, 1e-15)
        bin_mat = mat > 0
        wmax = bin_mat.sum(axis=0).flatten().max()
        threshold = (mu * max(1.0, wmax)) ** -(alpha + 1)

        # TPM expansion/inflation loop
        tpm = prune_matrix(tpm, threshold)
        last_tpm = np.zeros_like(tpm)
        while not np.allclose(tpm, last_tpm, rtol=rtol, atol=atol) and i < max_iter:
            last_tpm = tpm.copy()

            tpm = np.linalg.matrix_power(tpm, alpha)  # Expansion step
            tpm = np.power(tpm, beta)  # Inflation step
            tpm = prune_matrix(tpm, threshold)  # Threshold step

            i += 1

    # Cluster IDs are row numbers of max values in columns.
    cluster_ids = tpm.argmax(0)
    cluster_roots = create_clusters(graph, leafs, cluster_ids)  # type: ignore

    return graph, cluster_roots


def calculate_tpm(matrix: np.ndarray, mu: float) -> np.ndarray:
    """Calculate Transfer Probability Matrix (TPM), which in turn is the sum of the
    Relative Influence and Relative Depencency Matrices (RIM, RDM).

    Arguments:
        matrix: Adjacency matrix.
        mu: Decay constant.

    Returns:
        Transfer Probability Matrix (TPM).
    """
    relative = MarkovRelative(matrix, mu)

    # Relative Influence Matrix (RIM). Percentage of influence originating from j on i.
    rim = relative.influence_matrix

    # Relative Dependency Matrix (RDM). Percentage of dependency of j on i.
    rdm = relative.dependency_matrix

    # Create Transfer Probability Matrix (TPM).
    tpm = rim + rdm

    # Set diagonal to maximum nonzero column entry. This enforces a suitable aperiodic
    # stochastic matrix.
    np.fill_diagonal(tpm, 0.0)
    colmax = tpm.max(axis=0)
    colmax[colmax == 0.0] = 1.0
    np.fill_diagonal(tpm, colmax)

    # Normalize each column for a column sum of 1.
    tpm = tpm / np.sum(tpm, 0)

    return tpm


def prune_matrix(matrix: np.ndarray, threshold: float) -> np.ndarray:
    """Return a column normalized matrix for which all values below the threshold have
    been pruned (set to 0.0).

    Arguments:
        matrix: Matrix to prune.
        threshold: Cut-off threshold for normalized values.

    Returns:
        Pruned matrix.
    """
    colsum = matrix.sum(0)
    to_prune = np.logical_and(matrix < threshold * colsum, matrix > 0)

    while to_prune.any():
        matrix[to_prune] = 0
        colsum = matrix.sum(0)
        to_prune = np.logical_and(matrix < threshold * colsum, matrix > 0)

    matrix = matrix / colsum
    return matrix


def create_clusters(graph: Graph, nodes: List[Node], cluster_ids: np.ndarray) -> List[Node]:
    """Assign nodes in graph to new cluster nodes using a numbered array.

    Arguments:
        graph: Graph to add clusters into.
        nodes: Nodes that were clustered.
        cluster_ids: 1D array with numbered cluster assignment per node.

    Returns:
        Current root nodes in the graph.
    """
    assert (
        len(nodes) == cluster_ids.size
    ), "Node count should match cluster IDs length. Found %d vs %d." % (
        len(nodes),
        cluster_ids.size,
    )

    # Transform cluster_ids to ranks of their unique values.
    # This makes them a perfect index sequence starting at 0.
    unique_ids, id_ranks = np.unique(cluster_ids, return_inverse=True)

    # Prepare children lists and fill them.
    children_lists: List[List[Node]] = [[] for i in unique_ids]
    for i, node in enumerate(nodes):
        rank = id_ranks[i]
        children_lists[rank].append(node)

    # Create cluster nodes and keep track of cluster roots.
    new_cluster_roots = []
    unchanged_cluster_roots = []
    for children in children_lists:
        if len(children) > 1:
            parent = create_parent(graph, children)
            new_cluster_roots.append(parent)
        # Node is it's own parent, don't change anything.
        else:
            unchanged_cluster_roots.append(children[0])

    return new_cluster_roots + unchanged_cluster_roots


class MarkovFlow:
    """Results of Markov steady-state flow analysis.

    Arguments:
        matrix: Adjacency matrix (non-negative, IR/FAD convention).
        mu: Evaporation constant (>1.0).

    Note:
        Solves the flow balance equation: A_s @ f + f_in = f, for which:
            * A_s: normalized adjacency matrix with added evaporation sink.
            * f_in: injection vector (1.0 for every node except sink).
            * f = inv(I - A_s) @ f_in (flow vector)
            * Q = inv(I - A_s) (sensitivity matrix)
            * f = Q @ f_in
            * F = A_s * f.T (edge flow matrix)
    """

    def __init__(self, matrix: np.ndarray, mu: float, scale: bool):
        self.matrix = matrix * (matrix > 0)
        self.mu = mu
        self.scale = scale

    @property
    def dim(self) -> int:
        """Adjacency matrix dimension."""
        return len(self.matrix)

    @property
    def sink_matrix(self) -> np.ndarray:
        """Column normalized adjacency matrix with an added evaporation sink."""
        return get_sink_matrix(self.matrix, self.mu)

    @property
    def sensitivity_matrix(self) -> np.ndarray:
        """Sensitivity matrix (Q) with respect to input flow. f = Q @ f_in."""
        try:
            mat = np.eye(self.dim + 1) - self.sink_matrix
            return np.linalg.inv(mat)
        except LinAlgError:
            markov_analysis.log(
                f"Error when trying to invert matrix of dim {self.dim}:\n"
                + f"{mat}\n\nCorresponding adjacency matrix:\n{self.matrix}."
            )
            raise

    @property
    def in_vector(self) -> np.ndarray:
        """Inflow vector used normally. 1.0 inflow at every node except the sink."""
        if self.scale:
            f_in = np.append(self.matrix.sum(axis=0), 0).reshape(self.dim + 1, 1)
        else:
            f_in = np.ones((self.dim + 1, 1))
            f_in[-1, 0] = 0
        return f_in

    @property
    def flow_vector(self) -> np.ndarray:
        """Flow through every node if system is injected with
        [`self.in_vector`][ragraph.analysis.cluster._markov.MarkovFlow.in_vector].
        """
        return self.sensitivity_matrix @ self.in_vector

    @property
    def flow_matrix(self) -> np.ndarray:
        """Flow over every edge if nodal flow equal
        [`self.flow_vector`][ragraph.analysis.cluster._markov.MarkovFlow.flow_vector].
        """
        return self.sink_matrix * self.flow_vector.T


class MarkovRelative:
    """Markov relative influence and dependency calculations.

    Arguments:
        adj: Adjacency matrix (non-negative, IR/FAD convention).
        mu: Evaporation constant (>1.0).

    Note:
        Solves specific cases of the Markov flow analysis where each node is injected
        with a flow of 1 to calculate the relative influence. The network is also
        reversed and injected again, which results in relative dependency.
    """

    def __init__(self, adj: np.ndarray, mu: float):
        self.adj = adj
        self.mu = mu

    @property
    def influence_matrix(self) -> np.ndarray:
        """Relative influence matrix (RIM).
        Percentage of influence originating from j on i.
        """
        return MarkovFlow(self.adj, self.mu, False).sensitivity_matrix[:-1, :-1]

    @property
    def dependency_matrix(self) -> np.ndarray:
        """Relative Dependency Matrix (RDM). Percentage of dependency of j on i."""
        return MarkovFlow(self.adj.T, self.mu, False).sensitivity_matrix[:-1, :-1].T


def get_sink_matrix(matrix: np.ndarray, mu: float) -> np.ndarray:
    """Calculate a normalized flow distribution matrix with an evaporation sink node.

    Arguments:
        matrix: Adjacency matrix.
        mu: Evaporation constant.

    Returns:
        Normalized flow distribution matrix with an added evaporation sink node at the
            bottom right.

    Note:
        Matrix structure of:
        [a00 a01 ... a0n 0]
        [a10 a11 ... a1n 0]
        [... ... ... ... .]
        [an0 an1 ... ann 0]
        [ e   e  ...  e  0]
        Where all columns are normalized [0.0, 1.0]
    """
    cont = 1 / mu
    evap = 1 - cont
    dim = len(matrix)
    S = matrix.sum(axis=0)  # Column sums.
    E = evap * np.ones((1, dim))  # Evaporation row.

    # Fixup edges with no outputs
    no_outputs = S == 0
    S[no_outputs] = 1.0
    E[:, no_outputs] = 1.0

    A_n = cont * matrix / S  # Normalized flow matrix [0,cont].

    # Add evaporation sink to adjacency matrix.
    # A_s = [[A_mu, 0], [evap, 0]]
    A_s = np.block([[A_n, np.zeros((dim, 1))], [E, np.zeros((1, 1))]])
    return A_s



"""# 8 node directed graph example from Shaja and Sudhakar.

Reference:
    Shaja, A. S., & Sudhakar, K. (2010). Optimized sequencing of analysis components in
    multidisciplinary systems. Research in Engineering Design, 21(3), 173-187.
    [DOI: 10.1007/s00163-009-0082-5](https://doi.org/10.1007/s00163-009-0082-5)
"""

edge_weights = ["adjacency"]



"""# Elephant Specification Language datasets"""
import importlib
from pathlib import Path
from typing import List

HERE = Path(__file__).parent


def enum() -> List[str]:
    """Enumerate all available ESL datasets."""
    return sorted(
        [
            d.name
            for d in HERE.iterdir()
            if d.is_dir() and not d.name.startswith(".") and not d.name.startswith("_")
        ]
    )


def check(name: str):
    """Check whether a dataset exists."""
    available = enum()
    if name not in available:
        raise ValueError(
            "Dataset {} cannot be found. Please pick one of {}.".format(name, available)
        )


def info(name: str) -> str:
    """Get information about a dataset."""
    check(name)
    mod = importlib.import_module("ragraph.datasets.esl.{}".format(name))
    doc = mod.__doc__
    return str(doc)


def get(name: str):
    """Get a dataset.

    Arguments:
        name: Name of the dataset to get (see `ragraph.datasets.esl.enum()`).
    """
    from ragraph.io.esl import from_esl

    check(name)

    esl_files = [str(p) for p in (HERE / name).glob("*.esl")]

    graph = from_esl(*esl_files)
    graph.kind = name

    return graph



"""# Jaccard Similarity Index

The index compares two objects, and is calculated as the size of the overlap in properties divided
by total size of properties they possess.

For examples on 'object description functions', please refer to [`the similarity
utilities`][ragraph.analysis.similarity.utils].

References:
    Kosub, S. (2016). A note on the triangle inequality for the Jaccard distance. Retrieved from
    [arXiv.org](https://arxiv.org/pdf/1612.02696.pdf) Jaccard, P. (1901). Étude comparative de la
    distribution florale dans une portion des Alpes et du Jura. Bulletin de La Société Vaudoise Des
    Sciences Naturelles. [DOI: 10.5169/seals-266450](https://doi.org/10.5169/seals-266450)
"""

from typing import Any, Callable, List

import numpy as np


def _calculate(props1: np.array, props2: np.array) -> float:
    """Calculate the Jaccard Index by the boolean object description arrays."""
    both = np.logical_and(props1, props2).sum()
    either = np.logical_or(props1, props2).sum()
    if either:
        return both / either
    return 0.0


def jaccard_index(obj1: Any, obj2: Any, on: Callable[[Any], List[bool]]) -> float:
    """Calculate the Jaccard Similarity Index between to objects based on an object
    description function.

    Arguments:
        obj1: First object to compare.
        obj2: Second object to compare.
        on: Callable that takes an object and describes it with a list of booleans.
            Each entry indicates the possession of a property.

    Returns:
        Jaccard Similarity between two objects, which is calculated as the size of the
        overlap in properties divided by total size of properties they posess.
    """
    props1 = np.array(on(obj1))
    props2 = np.array(on(obj2))
    return _calculate(props1, props2)


def jaccard_matrix(objects: List[Any], on: Callable[[Any], List[bool]]) -> np.ndarray:
    """Calculate the Jaccard Similarity Index for a set of objects based on an object
    description function.

    Arguments:
        objects: List of objects to generate a similarity matrix for.
        on: Callable that takes an object and describes it with a list of booleans.
            Each entry indicates the possession of a property.
    """
    dim = len(objects)
    mapping = mapping_matrix(objects, on)

    matrix = np.eye(dim, dtype=float)
    for i, obj_i in enumerate(objects):
        for j, obj_j in enumerate(objects):
            if j <= i:
                continue
            value = _calculate(mapping[i, :], mapping[j, :])
            matrix[i, j] = value
            matrix[j, i] = value
    return matrix


def mapping_matrix(objects: List[Any], on: Callable[[Any], List[bool]]) -> np.ndarray:
    """Calculate an object-property mapping matrix where each entry (i,j) indicates the
    possession of property j by object i.

    Arguments:
        objects: List of objects to describe.
        on: Callable that takes an object and describes it with a list of booleans.
            Each entry indicates the possession of a property.
    """
    return np.array([on(obj) for obj in objects])



"""# RaGraph built-in datasets"""

import importlib
from pathlib import Path
from typing import List

from ragraph.graph import Graph
from ragraph.io.csv import from_csv

HERE = Path(__file__).parent


def enum() -> List[str]:
    """Enumerate all available graphs in datasets."""
    return sorted(
        [
            d.name
            for d in HERE.iterdir()
            if d.is_dir()
            and not d.name.startswith(".")
            and not d.name.startswith("_")
            and not d.name == "esl"
        ]
    )


def check(name: str):
    """Check whether a dataset exists."""
    available = enum()
    if name not in available:
        raise ValueError(
            "Dataset {} cannot be found. Please pick one of {}.".format(name, available)
        )


def info(name: str) -> str:
    """Get information about a dataset."""
    check(name)
    mod = importlib.import_module("ragraph.datasets.{}".format(name))
    doc = mod.__doc__
    return str(doc)


def get(name: str) -> Graph:
    """Get a dataset."""
    check(name)

    mod = importlib.import_module("ragraph.datasets.{}".format(name))
    nodes_path = HERE / name / (name + "_nodes.csv")
    edges_path = HERE / name / (name + "_edges.csv")

    node_weights = getattr(mod, "node_weights", None)
    edge_weights = getattr(mod, "edge_weights", None)
    return from_csv(
        nodes_path=nodes_path,
        edges_path=edges_path,
        csv_delimiter=";",
        iter_delimiter=";",
        node_weights=node_weights,
        edge_weights=edge_weights,
        name=name,
        kind="dataset",
    )



"""# Labels list plot component"""
from copy import deepcopy
from typing import List

from ragraph.node import Node
from ragraph.plot.generic import Component, Style


class Labels(Component):
    """Labels list plot component.

    Arguments:
        leafs: List of leaf nodes.
        style: Plot style mapping.
    """

    def __init__(self, leafs: List[Node], style: Style = Style()):
        labels = [node.name for node in leafs]
        if style.labels.shorten is True:
            short_labels = [label.split(".")[-1] for label in labels]
        elif style.labels.shorten:
            short_labels = [style.labels.shorten(label) for label in labels]
        else:
            short_labels = labels

        label_length = max(len(label) for label in short_labels) if len(short_labels) else 0
        fontsize = (
            int(0.6 * style.boxsize) if style.labels.fontsize is None else style.labels.fontsize
        )

        if style.labels.textorientation == "vertical":
            height = (
                label_length * fontsize * style.labels.fontaspectratio + 0.3 * style.boxsize
            )  # Text width plus 0.3 boxsize as margin.

            num = len(leafs)
            xmax = num * style.xstep
            width = xmax * style.boxsize

            xdata = [(i + 0.5) * style.xstep for i in range(num)]
            ydata = num * [-0.5 * height / style.boxsize]

            annotations = [
                dict(
                    x=x,
                    y=y,
                    text=text,
                    showarrow=False,
                    yshift=0,
                    textangle=-90,
                    yanchor="bottom",
                    font=dict(
                        color=style.labels.fontcolor,
                        family=style.labels.fontfamily,
                        size=fontsize,
                    ),
                )
                for x, y, text in zip(xdata, ydata, short_labels)
            ]

            xaxis, yaxis = deepcopy(style.labels.xaxis), deepcopy(style.labels.yaxis)
            xaxis.update(range=[0, num])
            yaxis.update(
                range=[-0.5 * height / style.boxsize, 0.5 * height / style.boxsize],
                scaleanchor="x",
                scaleratio=1.0,
            )
        else:
            # Default option.
            width = (
                label_length * fontsize * style.labels.fontaspectratio + 0.5 * style.boxsize
            )  # Text width plus 0.3 boxsize as margin.

            num = len(leafs)
            ymax = num * style.ystep
            height = ymax * style.boxsize

            xdata = num * [0.5 * width / style.boxsize]
            ydata = [ymax - (i + 0.5) * style.ystep for i in range(num)]

            annotations = [
                dict(
                    x=x,
                    y=y,
                    text=text,
                    showarrow=False,
                    yshift=0,
                    textangle=0,
                    xanchor="right",
                    font=dict(
                        color=style.labels.fontcolor,
                        family=style.labels.fontfamily,
                        size=fontsize,
                    ),
                )
                for x, y, text in zip(xdata, ydata, short_labels)
            ]

            xaxis, yaxis = deepcopy(style.labels.xaxis), deepcopy(style.labels.yaxis)
            xaxis.update(
                range=[-0.5 * width / style.boxsize, 0.5 * width / style.boxsize],
                scaleanchor="y",
                scaleratio=1.0,
            )
            yaxis.update(range=[0, num])

        super().__init__(
            width=width,
            height=height,
            traces=[],
            annotations=annotations,
            shapes=None,
            xaxis=xaxis,
            yaxis=yaxis,
        )



"""
# Gamma bus detection

References:
    Wilschut, T., Etman, L. F. P., Rooda, J. E., & Adan, I. J. B. F. (2017). Multilevel
    Flow-Based Markov Clustering for Design Structure Matrices. Journal of Mechanical
    Design, 139(12), 121402. [DOI: 10.1115/1.4037626](https://doi.org/10.1115/1.4037626)
"""

from typing import List, Optional, Tuple, Union

import numpy as np

from ragraph.analysis._classes import Bound, BusAnalysis, Parameter
from ragraph.graph import Graph, Node

gamma_params = {
    p.name: p
    for p in [
        Parameter(
            "gamma",
            float,
            "Threshold factor for bus nodes with respect to the median of the "
            "(nonzero) node degree distribution.",
            default=2.0,
            lower=Bound(1.0, inclusive=True, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        )
    ]
}

gamma_analysis = BusAnalysis(
    name="Gamma bus detection",
    description="Detect bus nodes by selecting node degree outliers by some factor "
    + "gamma w.r.t. median of nonzero node degrees.",
    parameters=gamma_params,
)


@gamma_analysis
def gamma_bus_detection(
    graph: Graph,
    root: Optional[Union[Node, str]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    names: bool = False,
    gamma: float = gamma_params["gamma"].default,  # type: ignore
    safe: bool = True,
    **kwargs,
) -> Union[Tuple[List[Node], List[Node]], Tuple[List[str], List[str]]]:
    """Detect bus nodes in a graph.

    Arguments:
        graph: Graph to detect bus nodes in.
        root: Root node of this bus detection analysis.
        leafs: Optional list of nodes to consider leafs during this bus detection cycle.
        inherit: Whether edges between descendants of nodes should be taken into account
            (if applicable).
        loops: Whether self-loop edges should be taken into account (if applicable).
        gamma: Bus threshold w.r.t. median of nonzero node degrees.
        names: Whether to return node names or node instances.

    Returns:
        Bus leafs.
        Nonbus leafs.

    Note:
        Works by selecting node degree outliers by some factor gamma w.r.t. median of
        nonzero node degrees.

    Reference:
        Wilschut, T., Etman, L. F. P., Rooda, J. E., & Adan, I. J. B. F. (2017). Multilevel
        Flow-Based Markov Clustering for Design Structure Matrices. Journal of Mechanical
        Design, 139(12), 121402. https://doi.org/10.1115/1.4037626
    """
    matrix = graph.get_adjacency_matrix(
        nodes=leafs, inherit=inherit, loops=loops, only=edge_weights
    )
    assert isinstance(matrix, np.ndarray)
    assert leafs is not None
    dim = matrix.shape[0]

    # Calculate interface matrix and node degrees.
    interfaces = matrix > 0
    degrees = interfaces.sum(0) + interfaces.sum(1)

    # Initialize arrays and enter heuristic.
    nonbus_idxs: np.ndarray = np.arange(dim)
    bus_idxs: np.ndarray = np.int_([])
    while nonbus_idxs.size > 1:
        nonbus_nonzero_degrees = degrees[nonbus_idxs][degrees[nonbus_idxs] > 0]

        # Calculate threshold if nonzero degree nonbus nodes are left.
        if not nonbus_nonzero_degrees.size:
            break
        threshold = gamma * np.median(nonbus_nonzero_degrees)

        # Transfer nodes to bus if found.
        add_bus = nonbus_idxs[degrees[nonbus_idxs] >= threshold]
        if not add_bus.size:
            break
        setdiff = np.setdiff1d(nonbus_idxs, add_bus, assume_unique=True)
        bus_idxs = np.hstack((bus_idxs, add_bus))
        nonbus_idxs = setdiff

    # If all nodes were assigned to the bus, actually none are.
    if len(bus_idxs) != dim and len(bus_idxs) > 0:
        bus_leafs = [leafs[i] for i in bus_idxs]
        nonbus_leafs = [leafs[i] for i in nonbus_idxs]
    else:
        bus_leafs = []
        nonbus_leafs = leafs

    return bus_leafs, nonbus_leafs



"""# LED system-in-package (LEDSiP, 1200 components)

Reference:
Wilschut, T. (2014). Analysis of the multi-disciplinary coupling structure of a LED
system-in-package.
"""

edge_weights = ["adjacency"]
